from mau.environment.environment import Environment
from mau.nodes.block import BlockGroupNode, BlockNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.paragraph import ParagraphNode
from mau.visitors.jinja_visitor import JinjaVisitor


def test_page_block_node_standard_block_template():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "block.j2": (
            "{{ subtype }} - {{ content }} - {{ secondary_content }} - "
            "{{ classes | join(',') }} - {{ title }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = BlockNode(
        subtype="someblock",
        children=[TextNode("my content")],
        secondary_children=[TextNode("my secondary content")],
        classes=["class1", "class2"],
        title=SentenceNode(children=[TextNode("sometitle")]),
        engine="someengine",
        preprocessor="somepreprocessor",
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == (
        "someblock - my content - my secondary content - class1,class2 - "
        "sometitle - arg1,arg2 - key1:value1 - tag1,tag2"
    )


def test_page_block_node_subtype_template_has_precedence():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "block.someblock.j2": "The subtype template",
        "block.j2": "The wrong template",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = BlockNode(
        subtype="someblock",
        children=[TextNode("my content")],
        secondary_children=[TextNode("my secondary content")],
        classes=["class1", "class2"],
        title=SentenceNode(children=[TextNode("sometitle")]),
        engine="someengine",
        preprocessor="somepreprocessor",
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == "The subtype template"


def test_page_block_node_engine_template_has_precedence():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "block.someengine.j2": "The engine template",
        "block.someblock.j2": "The wrong template",
        "block.j2": "The wrong template",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = BlockNode(
        subtype="someblock",
        children=[TextNode("my content")],
        secondary_children=[TextNode("my secondary content")],
        classes=["class1", "class2"],
        title=SentenceNode(children=[TextNode("sometitle")]),
        engine="someengine",
        preprocessor="somepreprocessor",
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == "The engine template"


def test_page_block_node_engine_and_subtype_template_has_precedence():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "block.someengine.someblock.j2": "The engine+block template",
        "block.someengine.j2": "The wrong template",
        "block.someblock.j2": "The wrong template",
        "block.j2": "The wrong template",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = BlockNode(
        subtype="someblock",
        children=[TextNode("my content")],
        secondary_children=[TextNode("my secondary content")],
        classes=["class1", "class2"],
        title=SentenceNode(children=[TextNode("sometitle")]),
        engine="someengine",
        preprocessor="somepreprocessor",
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == "The engine+block template"


def test_block_group():
    templates = {
        "text.j2": "{{ value }}",
        "paragraph.j2": "{{ content }}",
        "sentence.j2": "{{ content }}",
        "block.group.j2": "The block template {{ subtype }}",
        "block_group.somegroup.j2": "{% for block in group.values() %}{{ block }}-{% endfor %}",
        "block.j2": "The wrong template",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = BlockGroupNode(
        group_name="somegroup",
        group={
            "left": BlockNode(
                subtype="sometype1",
                children=[
                    ParagraphNode(
                        children=[
                            TextNode("Block 1"),
                        ]
                    ),
                ],
                secondary_children=[],
                classes=[],
                title=None,
                engine="group",
                preprocessor="none",
                args=[],
                kwargs={},
            ),
            "right": BlockNode(
                subtype="sometype2",
                children=[
                    ParagraphNode(
                        children=[
                            TextNode("Block 2"),
                        ]
                    ),
                ],
                secondary_children=[],
                classes=[],
                title=None,
                engine="group",
                preprocessor="none",
                args=[],
                kwargs={},
            ),
        },
        args=[],
        kwargs={},
        tags=[],
    )

    result = visitor.visit(node)

    assert result == "The block template sometype1-The block template sometype2-"



# pylint: disable=too-many-lines

from mau.environment.environment import Environment
from mau.lexers.base_lexer import TokenTypes as BLTokenTypes
from mau.lexers.main_lexer import MainLexer
from mau.lexers.main_lexer import TokenTypes as MLTokenTypes
from mau.nodes.block import BlockGroupNode, BlockNode
from mau.nodes.content import ContentImageNode, ContentNode
from mau.nodes.header import HeaderNode
from mau.nodes.inline import RawNode, SentenceNode
from mau.nodes.lists import ListItemNode, ListNode
from mau.nodes.page import ContainerNode, HorizontalRuleNode
from mau.nodes.paragraph import ParagraphNode
from mau.nodes.source import CalloutNode, CalloutsEntryNode, SourceNode
from mau.parsers.arguments_parser import ArgumentsParser
from mau.parsers.attributes import AttributesManager
from mau.parsers.base_parser import BaseParser
from mau.parsers.footnotes import FootnotesManager
from mau.parsers.internal_links import InternalLinksManager
from mau.parsers.preprocess_variables_parser import PreprocessVariablesParser
from mau.parsers.text_parser import TextParser
from mau.parsers.toc import TocManager, header_anchor
from mau.tokens.tokens import Token


# The MainParser is in charge of parsing
# the whole input, calling other parsers
# to manage single paragraphs or other
# things like variables.
class MainParser(BaseParser):
    lexer_class = MainLexer

    def __init__(self, environment, parent_node=None, parent_position=None):
        super().__init__(environment, parent_node, parent_position)

        self.internal_links_manager = InternalLinksManager(self)
        self.footnotes_manager = FootnotesManager(self)
        self.toc_manager = TocManager(self)
        self.attributes_manager = AttributesManager(self)

        # These are the default block aliases
        # If subtype is not set it will be the alias itself.
        self.block_aliases = {
            "source": {
                "subtype": None,
                "mandatory_args": ["language"],
                "defaults": {"engine": "source", "language": "text"},
            },
            "footnote": {
                "subtype": None,
                "mandatory_args": ["name"],
                "defaults": {"engine": "footnote"},
            },
            "admonition": {
                "mandatory_args": ["class", "icon"],
            },
        }

        # Iterate through block definitions passed as variables
        self.block_aliases.update(
            self.environment.getvar(
                "mau.parser.block_definitions", Environment()
            ).asdict()
        )

        # This is a buffer for a block title
        self.title = (None, None)

        # This is a buffer for a control
        self.control = (None, None, None)

        # This is the function used to create the header
        # anchors.
        self.header_anchor = self.environment.getvar(
            "mau.parser.header_anchor_function", header_anchor
        )

        # The last index in the latest ordered list,
        # used to calculate the beginning value of them
        # next one when start=auto
        self.latest_ordered_list_index = 0

        # This is the dictionary of block groups
        # defineed in the document
        self.grouped_blocks = {}

        # This is the final output of the parser
        self.output = {}

    def _process_functions(self):
        # All the functions that this parser provides.

        return [
            self._process_eol,
            self._process_horizontal_rule,
            self._process_single_line_comment,
            self._process_multi_line_comment,
            self._parse_variable_definition,
            self._process_command,
            self._process_title,
            self._process_control,
            self._process_arguments,
            self._process_header,
            self._process_block,
            self._process_content,
            self._process_list,
            self._process_paragraph,
        ]

    def _push_title(self, text, context):
        # When we parse a title we can store it here
        # so that it is available to the next block
        # that will use it.
        self.title = (text, context)

    def _pop_title(self, node):
        # This return the title and resets the
        # cached one, so no other block will
        # use it.
        text, context = self.title
        self._reset_title()

        if text is None:
            return None

        text_parser = TextParser.analyse(
            text,
            context,
            self.environment,
            parent_node=node,
            parent_position="title",
        )

        return SentenceNode(
            parent=node,
            parent_position="title",
            children=text_parser.nodes,
        )

    def _reset_title(self):
        self.title = (None, None)

    def _push_control(self, operator, statement, context):
        self.control = (operator, statement, context)

    def _pop_control(self):
        # This return the title and resets the
        # cached one, so no other block will
        # use it.
        operator, statement, context = self.control
        self._reset_control()

        if operator is None:
            return True

        if operator != "if":
            self._error(f"Control operator '{operator}' is not supported")

        try:
            variable, test = statement.split(":", 1)
        except ValueError:
            self._error(f"Statement '{statement}' is not in the form variable:test")

        variable_value = self.environment.getvar(variable, None)

        if variable_value is None:
            self._error(f"Variable '{variable}' has not been defined")

        if test.startswith("="):
            value = test[1:]
            return variable_value == value

        if test.startswith("!="):
            value = test[2:]

            return variable_value != value

        if test.startswith("&"):
            value = test[1:]

            if value not in ["true", "false"]:
                self._error(f"Boolean value '{value}' is invalid")

            # pylint: disable=simplifiable-if-expression
            value = True if value == "true" else False

            return variable_value and value

        self._error(f"Test '{test}' is not supported")

    def _reset_control(self):
        self.control = (None, None, None)

    def _collect_text_content(self):
        # Collects all adjacent text tokens
        # into a single string

        if not self._peek_token_is(BLTokenTypes.TEXT):  # pragma: no cover
            return None

        values = []

        # Get all tokens
        while self._peek_token_is(BLTokenTypes.TEXT):
            values.append(self._get_token().value)
            self._get_token(BLTokenTypes.EOL)

        return " ".join(values)

    def _parse_text_content(
        self, text, parent_node=None, parent_position=None, context=None
    ):
        # Parse a piece of text using the TextParser.

        current_context = context or self._current_token.context

        # Replace variables
        preprocess_parser = PreprocessVariablesParser.analyse(
            text,
            current_context,
            self.environment,
        )
        text = preprocess_parser.nodes[0].value

        # Parse the text
        text_parser = TextParser.analyse(
            text,
            current_context,
            self.environment,
            parent_node=parent_node,
            parent_position=parent_position,
        )

        # Extract the footnote mentions
        # found in this piece of text
        self.footnotes_manager.update_mentions(text_parser.footnotes)

        # Extract the internal links
        # found in this piece of text
        self.internal_links_manager.update_links(text_parser.links)

        return text_parser.nodes

    def _process_eol(self):
        # This simply parses the end of line.

        self._get_token(BLTokenTypes.EOL)

        return True

    def _process_horizontal_rule(self):
        # The horizontal rule ---

        self._get_token(MLTokenTypes.HORIZONTAL_RULE)

        args, kwargs, tags, subtype = self.attributes_manager.pop()

        self._save(
            HorizontalRuleNode(
                subtype=subtype,
                args=args,
                kwargs=kwargs,
                tags=tags,
            )
        )

        return True

    def _process_single_line_comment(self):
        # // A comment on a single line

        self._get_token(MLTokenTypes.COMMENT)
        self._get_token(BLTokenTypes.EOL)

        return True

    def _process_multi_line_comment(self):
        # ////
        # A comment
        # on multiple lines
        # ////

        self._get_token(MLTokenTypes.MULTILINE_COMMENT)
        self._collect([Token(MLTokenTypes.MULTILINE_COMMENT, "////")])
        self._force_token(MLTokenTypes.MULTILINE_COMMENT, "////")

        return True

    def _parse_variable_definition(self):
        # This parses the definition of a variable
        #
        # Simple variables are defined as :name:value
        # as True booleans as just :name:
        # and as False booleas as :!name:
        #
        # Variable names can use a namespace with
        # :namespace.name:value

        # Get the mandatory variable name
        self._get_token(MLTokenTypes.VARIABLE, ":")
        variable_token = self._get_token(BLTokenTypes.TEXT)
        self._get_token(BLTokenTypes.LITERAL, ":")

        context = variable_token.context
        variable_name = variable_token.value

        # Get the optional value
        value = self._collect_join([Token(BLTokenTypes.EOL)])

        # If the name starts with "+" it's a true flag
        # If the name starts with "-" it's a false flag
        if variable_name.startswith("+"):
            variable_name = variable_name[1:]
            value = True
        elif variable_name.startswith("-"):
            variable_name = variable_name[1:]
            value = False
        else:
            preprocess_parser = PreprocessVariablesParser.analyse(
                value,
                context,
                self.environment,
            )
            value = preprocess_parser.nodes[0].value

        self.environment.setvar(variable_name, value)

        return True

    def _process_command(self):
        # Parse a command in the form ::command:arguments

        self._get_token(MLTokenTypes.COMMAND, "::")
        name = self._get_token(BLTokenTypes.TEXT).value
        self._get_token(BLTokenTypes.LITERAL, ":")

        args, kwargs, tags, subtype = self.attributes_manager.pop()

        # Commands can have arguments
        command_args = []
        command_kwargs = {}
        with self:
            arguments = self._get_token(BLTokenTypes.TEXT)

            arguments_parser = ArgumentsParser.analyse(
                arguments.value, arguments.context, self.environment
            )

            (
                command_args,
                command_kwargs,
                _,
                command_subtype,
            ) = arguments_parser.process_arguments()

        if name == "defblock":
            if len(command_args) < 1:
                self._error("Block definitions require at least the alias")

            alias = command_args.pop(0)

            self.block_aliases[alias] = {
                "subtype": command_subtype,
                "mandatory_args": command_args,
                "defaults": command_kwargs,
            }

        elif name == "toc":
            self.toc_manager.create_toc_node(subtype, args, kwargs, tags)

        elif name == "footnotes":
            # Create a footnotes node
            self.footnotes_manager.create_node(subtype, args, kwargs, tags)

        elif name == "blockgroup":
            command_args, command_kwargs = self._set_names_and_defaults(
                command_args,
                command_kwargs,
                ["group"],
            )

            group_name = command_kwargs.pop("group")

            try:
                group = self.grouped_blocks.pop(group_name)
            except KeyError:
                self._error(
                    (
                        f"The group of blocks {group_name} doesn't exist. "
                        "No blocks belong to that group."
                    )
                )

            node = BlockGroupNode(
                group_name=group_name,
                group=group,
                subtype=subtype,
                args=args,
                kwargs=kwargs,
                tags=tags,
            )

            for position, block in group.items():
                block.parent = node
                block.parent_position = position

            self.save(node)

        return True

    def _process_title(self):
        # Parse a title in the form
        #
        # . This is a title
        # or
        # .This is a title

        # Parse the mandatory dot
        dot = self._get_token(MLTokenTypes.TITLE, ".")

        # Parse the optional white spaces
        with self:
            self._get_token(BLTokenTypes.WHITESPACE)

        # Get the text of the title
        text = self._get_token(BLTokenTypes.TEXT).value
        self._get_token(BLTokenTypes.EOL)

        self._push_title(text, dot.context)

        return True

    def _process_control(self):
        # Parse a control statement in the form
        #
        # @operator:control_statement

        # Parse the mandatory @
        at = self._get_token(MLTokenTypes.CONTROL, "@")

        # Get the operator
        operator = self._get_token(BLTokenTypes.TEXT).value

        # Discard the :
        self._get_token(BLTokenTypes.LITERAL, ":")

        # Get the statement
        statement = self._get_token(BLTokenTypes.TEXT).value

        self._get_token(BLTokenTypes.EOL)

        self._push_control(operator, statement, at.context)

        return True

    def _process_arguments(self):
        # Parse arguments in the form
        # [unnamed1, unnamed2, ..., named1=value1, name2=value2, ...]

        self._get_token(MLTokenTypes.ARGUMENTS, "[")
        text_token = self._get_token(BLTokenTypes.TEXT)
        self._get_token(BLTokenTypes.LITERAL, "]")

        preprocess_parser = PreprocessVariablesParser.analyse(
            text_token.value,
            text_token.context,
            self.environment,
        )
        text = preprocess_parser.nodes[0].value

        # Parse the text
        arguments_parser = ArgumentsParser.analyse(
            text, text_token.context, self.environment
        )

        args, kwargs, tags, subtype = arguments_parser.process_arguments()
        self.attributes_manager.push(args, kwargs, tags, subtype)

        return True

    def _process_header(self):
        # Parse a header in the form
        #
        # = Header
        #
        # The number of equal signs is arbitrary
        # and represents the level of the header.
        # Headers are automatically assigned an anchor
        # created using the provided function self.header_anchor

        # Get all the equal signs
        header = self._get_token(MLTokenTypes.HEADER).value

        # Get the mandatory white spaces
        self._get_token(BLTokenTypes.WHITESPACE)

        # Get the text of the header and calculate the level
        text_token = self._get_token(BLTokenTypes.TEXT)
        level = len(header)

        preprocess_parser = PreprocessVariablesParser.analyse(
            text_token.value,
            text_token.context,
            self.environment,
        )
        text = preprocess_parser.nodes[0].value

        # Check the control
        if self._pop_control() is False:
            return True

        # Consume the parser arguments
        args, kwargs, tags, subtype = self.attributes_manager.pop()

        # Create the anchor
        anchor = kwargs.pop("anchor", self.header_anchor(text, level))

        node = HeaderNode(
            level=str(level),
            anchor=anchor,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )

        # Titles can contain Mau code
        text_parser = TextParser.analyse(
            text,
            text_token.context,
            self.environment,
            parent_node=node,
        )
        node.value = SentenceNode(
            parent=node,
            children=text_parser.nodes,
        )

        # If there is an id store the header
        # to be processed by internal links
        if "id" in node.kwargs:
            self.internal_links_manager.add_header(node.kwargs["id"], node)

        self.toc_manager.add_header_node(node)

        self._save(node)

        return True

    def _collect_lines(self, stop_tokens):
        # This collects several lines of text in a list
        # until it gets to a line that begins with one
        # of the tokens listed in stop_tokens.
        # It is useful for block or other elements that
        # are clearly surrounded by delimiters.
        lines = []

        while self._peek_token() not in stop_tokens:
            lines.append(self._collect_join([Token(BLTokenTypes.EOL)]))
            self._get_token(BLTokenTypes.EOL)

        return lines

    def _process_block(
        self,
    ):  # pylint: disable=too-many-statements, too-many-branches, too-many-locals
        # Parse a block in the form
        #
        # [block_type]
        # ----
        # Content
        # ----
        # Optional secondary content
        #
        # Blocks are delimited by 4 consecutive identical characters.

        # Get the delimiter and check the length
        delimiter = self._get_token(MLTokenTypes.BLOCK).value

        self._get_token(BLTokenTypes.EOL)

        # Collect everything until the next delimiter
        content = self._collect_lines(
            [Token(MLTokenTypes.BLOCK, delimiter), Token(BLTokenTypes.EOF)]
        )

        self._force_token(MLTokenTypes.BLOCK, delimiter)
        self._get_token(BLTokenTypes.EOL)

        # Get the optional secondary content
        secondary_content = self._collect_lines(
            [Token(BLTokenTypes.EOL), Token(BLTokenTypes.EOF)]
        )

        if delimiter in secondary_content:
            # This probably means that the input contains an error
            # and we are in a situation like

            # ----
            #
            # ----
            # Text
            # ----

            # Where ["Text", "----"] is considered the secondary content
            # of an empty block.
            self._error(
                "Detected unclosed block (possibly before this line)"
            )  # pragma: no cover

        # Create the block
        block = BlockNode(children=content, secondary_children=secondary_content)

        block.title = self._pop_title(block)

        # Consume the arguments
        args, kwargs, tags, subtype = self.attributes_manager.pop()

        # Check the control
        if self._pop_control() is False:
            return True

        # If the subtype is an alias process it
        alias = self.block_aliases.get(subtype, {})
        block.subtype = alias.get("subtype", subtype)
        block_names = alias.get("mandatory_args", [])
        block_defaults = alias.get("defaults", {})

        args, kwargs = self._set_names_and_defaults(
            args,
            kwargs,
            block_names,
            block_defaults,
        )

        # Extract classes and convert them into a list
        classes = []
        if "classes" in kwargs:
            classes = kwargs.pop("classes")

            if classes:
                classes = classes.split(",")
        block.classes = classes

        # Extract the preprocessor
        block.preprocessor = kwargs.pop("preprocessor", "none")

        # Extract the engine
        block.engine = kwargs.pop("engine", None)

        block.args = args
        block.kwargs = kwargs
        block.tags = tags

        if block.engine is None:
            self._parse_default_engine(block)
        elif block.engine == "mau":
            self._parse_mau_engine(block)
        elif block.engine == "source":
            self._parse_source_engine(block)
        elif block.engine == "footnote":
            self._parse_footnote_engine(block)
        elif block.engine == "raw":
            self._parse_raw_engine(block)
        elif block.engine == "group":
            self._parse_group_engine(block)
        else:
            self._error(f"Engine {block.engine} is not available")

        return True

    def _parse_block_content(self, block):
        current_context = self._current_token.context

        content_parser = MainParser.analyse(
            "\n".join(block.children),
            current_context,
            Environment(),
            parent_node=block,
            parent_position="primary",
        )
        content_parser.finalise()

        secondary_content_parser = MainParser.analyse(
            "\n".join(block.secondary_children),
            current_context,
            Environment(),
            parent_node=block,
            parent_position="secondary",
        )
        secondary_content_parser.finalise()

        block.children = content_parser.nodes
        block.secondary_children = secondary_content_parser.nodes

    def _parse_block_content_update(self, block):
        current_context = self._current_token.context

        content_parser = MainParser.analyse(
            "\n".join(block.children),
            current_context,
            self.environment,
            parent_node=block,
            parent_position="primary",
        )

        secondary_content_parser = MainParser.analyse(
            "\n".join(block.secondary_children),
            current_context,
            self.environment,
            parent_node=block,
            parent_position="secondary",
        )

        block.children = content_parser.nodes
        block.secondary_children = secondary_content_parser.nodes

        # The footnote mentions and definitions
        # found in this block are part of the
        # main document. Import them.
        self.footnotes_manager.update(content_parser.footnotes_manager)

        # The internal links and headers
        # found in this block are part of the
        # main document. Import them.
        self.internal_links_manager.update(content_parser.internal_links_manager)

        self.toc_manager.update(content_parser.toc_manager)

    def _parse_mau_engine(self, block):
        self._parse_block_content(block)
        self._save(block)

    def _parse_default_engine(self, block):
        self._parse_block_content_update(block)
        self._save(block)

    def _parse_group_engine(self, block):
        block.args, block.kwargs = self._set_names_and_defaults(
            block.args,
            block.kwargs,
            ["group", "position"],
        )

        group_name = block.kwargs.pop("group")
        position = block.kwargs.pop("position")

        group = self.grouped_blocks.setdefault(group_name, {})

        if position in group:
            self._error(
                f"Block with position {position} already defined in group {group_name}"
            )

        group[position] = block

        self._parse_block_content_update(block)

    def _parse_footnote_engine(self, block):
        # The current block contains footnote data.
        # Extract the content and store it in
        # the footnotes manager.
        name = block.kwargs.pop("name")

        content_parser = MainParser.analyse(
            "\n".join(block.children),
            self._current_token.context,
            self.environment,
            parent_node=block,
        )

        self.footnotes_manager.add_data(name, content_parser.nodes)

    def _parse_raw_engine(self, block):
        # Engine "raw" doesn't process the content,
        # so we just pass it untouched in the form of
        # a RawNode per line.
        block.children = [RawNode(line) for line in block.children]
        block.secondary_children = [RawNode(line) for line in block.secondary_children]

        self._save(block)

    def _parse_source_engine(self, block):  # pylint: disable=too-many-locals
        # Parse a source block in the form
        #
        # [source, language, attributes...]
        # ----
        # content
        # ----
        #
        # Source blocks support the following attributes
        #
        # callouts=":" The separator used by callouts
        # highlight="@" The special character to turn on highlight
        #
        # [source, language, attributes...]
        # ----
        # content:1:
        # ----
        #
        # [source, language, attributes...]
        # ----
        # content:@:
        # ----
        #
        # Callout descriptions can be added to the block
        # as secondary content with the syntax
        #
        # [source, language, attributes...]
        # ----
        # content:name:
        # ----
        # <name>: <description>
        #
        # Since Mau uses Pygments, the attribute language
        # is one of the languages supported by that tool.

        # Get the delimiter for callouts (":" by default)
        delimiter = block.kwargs.pop("callouts", ":")

        # A list that contains callout markers in
        # the form (linenum,name)
        callout_markers = []

        # Get the marker for highlighted lines ("@" by default)
        highlight_marker = block.kwargs.pop("highlight", "@")

        # A list of highlighted lines
        highlighted_lines = []

        # Get the language
        language = block.kwargs.pop("language")

        # Source blocks preserve anything is inside

        # This is a list of all lines that might contain
        # a callout. They will be further processed
        # later to be sure.
        lines_with_callouts = [
            (linenum, line)
            for linenum, line in enumerate(block.children)
            if line.endswith(delimiter)
        ]

        # Each line in the previous list is processed
        # and stored if it contains a callout
        for linenum, line in lines_with_callouts:
            # Remove the final delimiter
            line = line[:-1]

            splits = line.split(delimiter)
            if len(splits) < 2:
                # It's a trap! There are no separators left
                continue

            # Get the callout and the line
            callout_name = splits[-1]
            line = delimiter.join(splits[:-1])

            block.children[linenum] = line

            # Check if we want to just highlight the line
            if callout_name == highlight_marker:
                highlighted_lines.append(linenum)
            else:
                callout_markers.append(CalloutNode(linenum, callout_name))

        # A list of CalloutEntryNode objects that contain the
        # text for each marker
        callout_contents = []

        # If there was secondary content it should be formatted
        # with callout names followed by colon and the
        # callout text.
        for line in block.secondary_children:
            if ":" not in line:
                self._error(
                    (
                        "Callout description should be written "
                        f"as 'name: text'. Missing ':' in '{line}'"
                    )
                )

            name, text = line.split(":")

            text = text.strip()

            callout_contents.append(CalloutsEntryNode(name, text))

        # Source blocks must preserve the content literally.
        # However, we need to remove escape characters from directives.
        # Directives are processed by the lexer, so if we want to
        # prevent Mau from interpreting them we have to escape them.
        # Escape characters are preserved by source blocks as anything
        # else, but in this case the character should be removed.
        textlines = []
        for line in block.children:
            if line.startswith(r"\::#"):
                line = line[1:]

            textlines.append(RawNode(line))

        node = SourceNode(
            code=textlines,
            language=language,
            callouts=callout_contents,
            highlights=highlighted_lines,
            markers=callout_markers,
            title=block.title,
            subtype=block.subtype,
            args=block.args,
            kwargs=block.kwargs,
            tags=block.tags,
        )

        self._save(node)

    def _process_content(self):
        # Parse content in the form
        #
        # << content_type:URI

        # Get the mandatory "<<"
        self._get_token(MLTokenTypes.CONTENT)

        with self:
            self._get_token(BLTokenTypes.WHITESPACE)

        # Get the content type
        content_type = self._get_token(BLTokenTypes.TEXT).value
        self._get_token(BLTokenTypes.LITERAL, ":")

        args, kwargs, tags, subtype = self.attributes_manager.pop()

        uris = self._get_token(BLTokenTypes.TEXT)

        with self:
            arguments_parser = ArgumentsParser.analyse(
                uris.value, uris.context, self.environment
            )

            uris, _, _, _ = arguments_parser.process_arguments()

        # Check the control
        if self._pop_control() is False:
            return True

        if content_type == "image":
            return self._parse_content_image(uris, subtype, args, kwargs, tags)

        return self._parse_standard_content(
            content_type, uris, subtype, args, kwargs, tags
        )

    def _parse_content_image(self, uris, subtype, args, kwargs, tags):
        # Parse a content image in the form
        #
        # << image:uri,alt_text,classes
        #
        # alt_text is the alternate text to use is the image is not reachable
        # and classes is a comma-separated list of classes

        # Consume the arguments
        args, kwargs = self._set_names_and_defaults(
            args,
            kwargs,
            ["alt_text", "classes"],
            {"alt_text": None, "classes": None},
        )

        uri = uris[0]
        alt_text = kwargs.pop("alt_text")
        classes = kwargs.pop("classes")

        if classes:
            classes = classes.split(",")

        node = ContentImageNode(
            uri=uri,
            alt_text=alt_text,
            classes=classes,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )

        node.title = self._pop_title(node)
        self._save(node)

        return True

    def _parse_standard_content(self, content_type, uris, subtype, args, kwargs, tags):
        # This is the fallback for an unknown content type

        node = ContentNode(
            content_type=content_type,
            uris=uris,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )

        node.title = self._pop_title(node)
        self._save(node)

        return True

    def _process_list(self):
        # Parse a list.
        # Lists can be ordered (using numbers)
        #
        # * One item
        # * Another item
        #
        # or unordered (using bullets)
        #
        # # Item 1
        # # Item 2
        #
        # The number of headers increases
        # the depth of each item
        #
        # # Item 1
        # ## Sub-Item 1.1
        #
        # Spaces before and after the header are ignored.
        # So the previous list can be also written
        #
        # # Item 1
        #   ## Sub-Item 1.1
        #
        # Ordered and unordered lists can be mixed.
        #
        # * One item
        # ## Sub Item 1
        # ## Sub Item 2
        #

        # Ignore initial white spaces
        with self:
            self._get_token(BLTokenTypes.WHITESPACE)

        # Get the header and decide if it's a numbered or unnumbered list
        header = self._peek_token(MLTokenTypes.LIST)
        ordered = header.value[0] == "#"

        args, kwargs, tags, subtype = self.attributes_manager.pop()

        # Parse all the following items
        nodes = self._process_list_nodes()

        if (start := kwargs.pop("start", 1)) == "auto":
            start = self.latest_ordered_list_index
            self.latest_ordered_list_index += len(nodes)
        else:
            start = int(start)
            self.latest_ordered_list_index = len(nodes) + start

        self._save(
            ListNode(
                ordered=ordered,
                main_node=True,
                start=start,
                children=nodes,
                subtype=subtype,
                args=args,
                kwargs=kwargs,
                tags=tags,
            )
        )

        return True

    def _process_list_nodes(self):
        # This parses all items of a list

        # Ignore initial white spaces
        with self:
            self._get_token(BLTokenTypes.WHITESPACE)

        # Parse the header and ignore the following white spaces
        header = self._get_token(MLTokenTypes.LIST).value
        self._get_token(BLTokenTypes.WHITESPACE)

        # Get the context of the first text token
        context = self._peek_token().context

        # Collect and parse the text of the item
        text = self._collect_text_content()
        content = self._parse_text_content(
            text,
            parent_node=self.parent_node,
            parent_position=self.parent_position,
            context=context,
        )

        # Compute the level of the item
        level = len(header)

        nodes = []
        nodes.append(ListItemNode(level=str(level), children=content))

        while self._peek_token() not in [
            Token(BLTokenTypes.EOF),
            Token(BLTokenTypes.EOL),
        ]:
            # This are the nodes inside the last element added to the list
            # which is used to append potential nested elements
            last_element_nodes = nodes[-1].children

            # Ignore the initial white spaces
            with self:
                self._get_token(BLTokenTypes.WHITESPACE)

            if len(self._peek_token().value) == level:
                # The new item is on the same level

                # Get the header
                header = self._get_token().value

                # Ignore white spaces
                self._get_token(BLTokenTypes.WHITESPACE)

                # Get the context of the first text token
                context = self._peek_token().context

                # Collect and parse the text of the item
                text = self._collect_text_content()
                content = self._parse_text_content(
                    text,
                    parent_node=self.parent_node,
                    parent_position=self.parent_position,
                    context=context,
                )

                level = len(header)

                nodes.append(ListItemNode(level=str(level), children=content))
            elif len(self._peek_token().value) > level:
                # The new item is on a deeper level

                # Treat the new line as a new list
                numbered = self._peek_token().value[0] == "#"
                subnodes = self._process_list_nodes()

                last_element_nodes.append(ListNode(ordered=numbered, children=subnodes))
            else:
                break

        return nodes

    def _process_paragraph(self):
        # This parses a paragraph.
        # Paragraphs can be written on multiple lines and
        # end with an empty line.

        # Get the context of the first token we are going to process
        context = self._peek_token().context

        # Each line ends with EOL. This collects everything
        # until the EOL, then removes it. If the next token
        # is EOL we know that the paragraph is ended, otherwise
        # we continue to collect. If the token is EOF we
        # reached the end and we have to stop anyway.
        lines = []
        while self._peek_token() not in [
            Token(BLTokenTypes.EOL),
            Token(BLTokenTypes.EOF),
        ]:
            lines.append(
                self._collect_join(
                    [
                        Token(BLTokenTypes.EOL),
                    ]
                )
            )
            self._get_token(BLTokenTypes.EOL)

        # Check the control
        if self._pop_control() is False:
            return True

        text = " ".join(lines)

        # Consume the arguments
        args, kwargs, tags, subtype = self.attributes_manager.pop()

        node = ParagraphNode(
            parent=self.parent_node,
            parent_position=self.parent_position,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )

        node.title = self._pop_title(node)

        node.children = self._parse_text_content(
            text, parent_node=node, context=context
        )

        self._save(node)

        return True

    def finalise(self):
        super().finalise()

        # This processes all footnotes stored in
        # the manager merging mentions and data
        # and updating the nodes that contain
        # a list of footnotes
        self.footnotes_manager.process_footnotes()

        # This processes all links stored in
        # the manager linking them to the
        # correct headers
        self.internal_links_manager.process_links()

        # Process ToC
        toc = self.toc_manager.process_toc()

        # The content wrappers are cloned to avoid
        # storing the whole parsed document in the
        # environment.
        content_wrapper = self.environment.getvar(
            "mau.parser.content_wrapper", ContainerNode()
        ).clone()
        content_wrapper.children = self.nodes

        toc_wrapper = self.environment.getvar(
            "mau.parser.toc_wrapper", ContainerNode()
        ).clone()
        toc_wrapper.add_children([toc])

        self.output.update(
            {
                "content": content_wrapper,
                "toc": toc_wrapper,
            }
        )



from mau.nodes.nodes import Node


class FootnoteNode(Node):
    node_type = "footnote"

    def __init__(
        self,
        number=None,
        reference_anchor=None,
        content_anchor=None,
        parent=None,
        parent_position=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.number = number
        self.reference_anchor = reference_anchor
        self.content_anchor = content_anchor

    def _custom_dict(self):
        return {
            "number": self.number,
            "reference_anchor": self.reference_anchor,
            "content_anchor": self.content_anchor,
        }

    def to_entry(self):
        return FootnotesEntryNode(
            number=self.number,
            reference_anchor=self.reference_anchor,
            content_anchor=self.content_anchor,
            parent=self.parent,
            parent_position=self.parent_position,
            children=self.children,
            subtype=self.subtype,
            args=self.args,
            kwargs=self.kwargs,
            tags=self.tags,
        )


class MacroFootnoteNode(FootnoteNode):
    """A footnote created inside a piece of text."""

    node_type = "footnote"


class FootnotesEntryNode(FootnoteNode):
    """An entry of the list of footnotes."""

    node_type = "footnotes_entry"


class FootnotesNode(Node):
    """This instructs Mau to insert the list of footnotes."""

    node_type = "footnotes"



from mau.lexers.text_lexer import TextLexer
from mau.nodes.inline import TextNode
from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(TextLexer, TextParser)

runner = parser_runner_factory(TextLexer, TextParser)


def test_empty_text():
    source = ""

    assert runner(source).nodes == []


def test_parse_word():
    source = "Word"

    expected = [
        TextNode("Word"),
    ]

    assert runner(source).nodes == expected


def test_multiple_words():
    source = "Many different words"

    expected = [
        TextNode("Many different words"),
    ]

    assert runner(source).nodes == expected


def test_parse_escape_word():
    source = r"\Escaped"

    expected = [
        TextNode("Escaped"),
    ]

    assert runner(source).nodes == expected


def test_parse_escape_symbol():
    source = r"\"Escaped"

    expected = [
        TextNode('"Escaped'),
    ]

    assert runner(source).nodes == expected


def test_square_brackets():
    source = "This contains [ and ] and [this]"

    expected = [
        TextNode("This contains [ and ] and [this]"),
    ]

    assert runner(source).nodes == expected



from mau.lexers.text_lexer import TextLexer
from mau.nodes.macros import MacroImageNode
from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(TextLexer, TextParser)

runner = parser_runner_factory(TextLexer, TextParser)


def test_macro_image():
    source = "[image](/the/path.jpg)"

    expected = [
        MacroImageNode("/the/path.jpg"),
    ]

    assert runner(source).nodes == expected


def test_macro_image_with_alt_text():
    source = '[image](/the/path.jpg,"alt name")'

    expected = [
        MacroImageNode("/the/path.jpg", alt_text="alt name"),
    ]

    assert runner(source).nodes == expected


def test_macro_image_with_width_and_height():
    source = "[image](/the/path.jpg,width=1200,height=600)"

    expected = [
        MacroImageNode("/the/path.jpg", alt_text=None, width="1200", height="600"),
    ]

    assert runner(source).nodes == expected



class MauErrorException(ValueError):
    """
    This is a processing error that can occour
    during lexing, parsing, or visiting.
    """

    def __init__(self, error):
        super().__init__(error.message)

        self.error = error


class MauError:
    """
    This is a processing error that can occour
    during lexing, parsing, or visiting.
    """

    source = "global"

    def __init__(self, message, details=None):
        super().__init__()

        self.message = message

        # These are specific values needed by
        # this error, e.g. the context for
        # errors related to tokens
        self.details = details or {}

    def print_details(self):  # pragma: no cover
        print(f"Error: {self.message}")


def print_error(error):  # pragma: no cover
    # This is a function used to print an error occurred
    # during processing.

    len_bar = 80
    title = f" {error.source} error "

    # This takes into account the title and the spaces
    half_bar = "=" * ((len_bar - len(title)) // 2 + 1)
    full_bar = half_bar + "=" * len(title) + half_bar

    print(f"{half_bar}{title}{half_bar}")

    error.print_details()

    print(f"{full_bar}")



from mau.nodes.nodes import Node


class TocEntryNode(Node):
    """An entry of the Table of Contents.

    This node contains an entry of the Table of Contents.
    """

    node_type = "toc_entry"

    def __init__(
        self,
        value=None,
        anchor=None,
        parent=None,
        parent_position=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.value = value
        self.anchor = anchor

    def _custom_dict(self):
        return {
            "value": self.value,
            "anchor": self.anchor,
        }


class TocNode(Node):
    """A Table of Contents command.

    This node contains the headers that go into the ToC.
    """

    node_type = "toc"



import textwrap
from unittest.mock import patch

import pytest
from mau.errors import MauErrorException
from mau.lexers.main_lexer import MainLexer
from mau.nodes.block import BlockNode
from mau.nodes.footnotes import FootnoteNode, FootnotesEntryNode, FootnotesNode
from mau.nodes.inline import TextNode
from mau.nodes.paragraph import ParagraphNode
from mau.parsers.footnotes import footnote_anchor
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_command_footnotes():
    source = """
    [*subtype1, arg1, arg2, #tag1, key1=value1, key2=value2]
    ::footnotes:
    """

    parser = runner(source)
    parser.footnotes_manager.footnotes = []

    assert parser.nodes == [
        FootnotesNode(
            children=parser.footnotes_manager.footnotes,
            subtype="subtype1",
            args=["arg1", "arg2"],
            kwargs={"key1": "value1", "key2": "value2"},
            tags=["tag1"],
        ),
    ]


@patch("mau.parsers.footnotes.hashlib.md5")
def test_default_footnote_anchor_function(mock_md5):
    mock_md5().hexdigest.return_value = "XXYYXXYYZZZ"

    assert footnote_anchor("Some Words 1234 56") == "XXYYXXYY"


def test_footnote_empty_block():
    source = """
    [*footnote, note1]
    ----
    ----
    """

    parser = runner(textwrap.dedent(source))
    parser.footnotes = {
        "note1": FootnoteNode(number=1),
    }

    assert parser.nodes == []


@patch("mau.parsers.footnotes.footnote_anchor")
def test_footnote_content(mock_footnote_anchor):
    mock_footnote_anchor.return_value = "XXYY"

    source = """
    [*footnote, note2]
    ----
    This is another paragraph.
    ----

    [*footnote, note1]
    ----
    This is a paragraph.
    ----
    """

    parser = runner(textwrap.dedent(source))
    parser.footnotes_manager.mentions = {
        "note1": FootnoteNode(number=1),
        "note2": FootnoteNode(number=2),
    }
    parser.finalise()

    assert parser.nodes == []
    assert parser.footnotes_manager.mentions == {
        "note1": FootnoteNode(
            number=1,
            children=[
                ParagraphNode(
                    children=[
                        TextNode("This is a paragraph."),
                    ]
                )
            ],
            reference_anchor="ref-footnote-1-XXYY",
            content_anchor="cnt-footnote-1-XXYY",
        ),
        "note2": FootnoteNode(
            number=2,
            children=[
                ParagraphNode(
                    children=[
                        TextNode("This is another paragraph."),
                    ]
                ),
            ],
            reference_anchor="ref-footnote-2-XXYY",
            content_anchor="cnt-footnote-2-XXYY",
        ),
    }


@patch("mau.parsers.footnotes.footnote_anchor")
def test_footnote_mention_and_content(mock_footnote_anchor):
    mock_footnote_anchor.return_value = "XXYY"

    source = """
    This is a paragraph with a footnote[footnote](note1).
    
    [*footnote, note1]
    ----
    This is the content of the footnote
    ----
    """

    parser = runner(textwrap.dedent(source))

    assert parser.footnotes_manager.mentions == {
        "note1": FootnoteNode(
            number=1,
            children=[
                ParagraphNode(
                    children=[
                        TextNode("This is the content of the footnote"),
                    ]
                ),
            ],
            reference_anchor="ref-footnote-1-XXYY",
            content_anchor="cnt-footnote-1-XXYY",
        )
    }
    assert parser.footnotes_manager.data == {
        "note1": [
            ParagraphNode(
                children=[
                    TextNode("This is the content of the footnote"),
                ]
            )
        ],
    }


@patch("mau.parsers.footnotes.footnote_anchor")
def test_footnotes_output(mock_footnote_anchor):
    mock_footnote_anchor.return_value = "XXYY"

    source = """
    This is a paragraph with a footnote[footnote](note1).
    
    [*footnote, note1]
    ----
    This is the content of the footnote
    ----

    ::footnotes:
    """

    parser = runner(textwrap.dedent(source))

    assert parser.nodes == [
        ParagraphNode(
            children=[
                TextNode("This is a paragraph with a footnote"),
                FootnoteNode(
                    children=[
                        ParagraphNode(
                            children=[
                                TextNode("This is the content of the footnote"),
                            ]
                        )
                    ],
                    number=1,
                    reference_anchor="ref-footnote-1-XXYY",
                    content_anchor="cnt-footnote-1-XXYY",
                ),
                TextNode("."),
            ]
        ),
        FootnotesNode(
            children=[
                FootnotesEntryNode(
                    children=[
                        ParagraphNode(
                            children=[
                                TextNode("This is the content of the footnote"),
                            ]
                        )
                    ],
                    number=1,
                    reference_anchor="ref-footnote-1-XXYY",
                    content_anchor="cnt-footnote-1-XXYY",
                ),
            ]
        ),
    ]


@patch("mau.parsers.footnotes.footnote_anchor")
def test_footnotes_output_in_block(mock_footnote_anchor):
    mock_footnote_anchor.return_value = "XXYY"

    source = """
    ++++
    This is a paragraph with a footnote[footnote](note1).
    
    [*footnote, note1]
    ----
    This is the content of the footnote
    ----

    ::footnotes:
    ++++
    """

    parser = runner(textwrap.dedent(source))

    assert parser.nodes == [
        BlockNode(
            children=[
                ParagraphNode(
                    children=[
                        TextNode("This is a paragraph with a footnote"),
                        FootnoteNode(
                            children=[
                                ParagraphNode(
                                    children=[
                                        TextNode("This is the content of the footnote"),
                                    ]
                                )
                            ],
                            number=1,
                            reference_anchor="ref-footnote-1-XXYY",
                            content_anchor="cnt-footnote-1-XXYY",
                        ),
                        TextNode("."),
                    ]
                ),
                FootnotesNode(
                    children=[
                        FootnotesEntryNode(
                            children=[
                                ParagraphNode(
                                    children=[
                                        TextNode("This is the content of the footnote"),
                                    ]
                                )
                            ],
                            number=1,
                            reference_anchor="ref-footnote-1-XXYY",
                            content_anchor="cnt-footnote-1-XXYY",
                        ),
                    ]
                ),
            ],
            preprocessor="none",
        )
    ]


def test_footnote_duplication():
    source = """
    [*footnote, note1]
    ----
    This is the content of the footnote
    ----

    This is a paragraph with a footnote[footnote](note1).

    This is a paragraph with the same footnote[footnote](note1).
    """

    with pytest.raises(MauErrorException):
        runner(textwrap.dedent(source))


@patch("mau.parsers.footnotes.footnote_anchor")
def test_footnote_data_and_mention_in_block(mock_footnote_anchor):
    mock_footnote_anchor.return_value = "XXYY"

    source = """
    ====
    This is a paragraph with a footnote[footnote](note1).

    [*footnote, note1]
    ----
    This is the content of the footnote
    ----
    ====
    """

    parser = runner(textwrap.dedent(source))

    assert parser.footnotes_manager.mentions == {
        "note1": FootnoteNode(
            number=1,
            children=[
                ParagraphNode(
                    children=[
                        TextNode("This is the content of the footnote"),
                    ]
                ),
            ],
            reference_anchor="ref-footnote-1-XXYY",
            content_anchor="cnt-footnote-1-XXYY",
        )
    }
    assert parser.footnotes_manager.data == {
        "note1": [
            ParagraphNode(
                children=[
                    TextNode("This is the content of the footnote"),
                ]
            )
        ],
    }


@patch("mau.parsers.footnotes.footnote_anchor")
def test_footnote_data_in_block(mock_footnote_anchor):
    mock_footnote_anchor.return_value = "XXYY"

    source = """
    This is a paragraph with a footnote[footnote](note1).

    ====
    [*footnote, note1]
    ----
    This is the content of the footnote
    ----
    ====
    """

    parser = runner(textwrap.dedent(source))

    assert parser.footnotes_manager.mentions == {
        "note1": FootnoteNode(
            number=1,
            children=[
                ParagraphNode(
                    children=[
                        TextNode("This is the content of the footnote"),
                    ]
                ),
            ],
            reference_anchor="ref-footnote-1-XXYY",
            content_anchor="cnt-footnote-1-XXYY",
        )
    }
    assert parser.footnotes_manager.data == {
        "note1": [
            ParagraphNode(
                children=[
                    TextNode("This is the content of the footnote"),
                ]
            )
        ],
    }


@patch("mau.parsers.footnotes.footnote_anchor")
def test_footnote_mention_in_block(mock_footnote_anchor):
    mock_footnote_anchor.return_value = "XXYY"

    source = """
    ====
    This is a paragraph with a footnote[footnote](note1).
    ====

    [*footnote, note1]
    ----
    This is the content of the footnote
    ----
    """

    parser = runner(textwrap.dedent(source))

    assert parser.footnotes_manager.mentions == {
        "note1": FootnoteNode(
            number=1,
            children=[
                ParagraphNode(
                    children=[
                        TextNode("This is the content of the footnote"),
                    ]
                ),
            ],
            reference_anchor="ref-footnote-1-XXYY",
            content_anchor="cnt-footnote-1-XXYY",
        )
    }
    assert parser.footnotes_manager.data == {
        "note1": [
            ParagraphNode(
                children=[
                    TextNode("This is the content of the footnote"),
                ]
            )
        ],
    }


# If footnotes are not assigned correctly,
# the following test will fail as the parent of
# both footnotes will be the second FootnotesNode
@patch("mau.parsers.footnotes.footnote_anchor")
def test_footnotes_parent(mock_footnote_anchor):
    mock_footnote_anchor.return_value = "XXYY"

    source = """
    Footnote [footnote](note1).

    [*footnote, note1]
    ----
    ----

    ::footnotes:

    [#tag1]
    ::footnotes:
    """

    parser = runner(textwrap.dedent(source))

    first_footnotes = parser.nodes[1]
    second_footnotes = parser.nodes[2]

    first_footnotes_footnote = first_footnotes.children[0]
    second_footnotes_footnote = second_footnotes.children[0]

    assert first_footnotes_footnote.parent == first_footnotes
    assert second_footnotes_footnote.parent == second_footnotes



from mau.environment.environment import Environment
from mau.lexers.main_lexer import MainLexer
from mau.nodes.block import BlockNode
from mau.nodes.header import HeaderNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.page import ContainerNode
from mau.nodes.toc import TocEntryNode, TocNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_command_toc():
    source = """
    [*subtype1, arg1, arg2, #tag1, key1=value1, key2=value2]
    ::toc:
    """

    assert runner(source).nodes == [
        TocNode(
            [],
            subtype="subtype1",
            args=["arg1", "arg2"],
            kwargs={"key1": "value1", "key2": "value2"},
            tags=["tag1"],
        ),
    ]


def test_toc():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: f"{text}-XXXXXX"
    )

    source = """
    = Header 1
    == Header 1.1
    = Header 2

    ::toc:
    """

    parser = runner(source, environment)

    assert parser.output == {
        "content": ContainerNode(
            children=[
                HeaderNode(
                    value=SentenceNode(children=[TextNode("Header 1")]),
                    level="1",
                    anchor="Header 1-XXXXXX",
                ),
                HeaderNode(
                    value=SentenceNode(children=[TextNode("Header 1.1")]),
                    level="2",
                    anchor="Header 1.1-XXXXXX",
                ),
                HeaderNode(
                    value=SentenceNode(children=[TextNode("Header 2")]),
                    level="1",
                    anchor="Header 2-XXXXXX",
                ),
                TocNode(
                    children=[
                        TocEntryNode(
                            value=SentenceNode(children=[TextNode("Header 1")]),
                            anchor="Header 1-XXXXXX",
                            children=[
                                TocEntryNode(
                                    value=SentenceNode(
                                        children=[TextNode("Header 1.1")]
                                    ),
                                    anchor="Header 1.1-XXXXXX",
                                    children=[],
                                ),
                            ],
                        ),
                        TocEntryNode(
                            value=SentenceNode(children=[TextNode("Header 2")]),
                            anchor="Header 2-XXXXXX",
                            children=[],
                        ),
                    ]
                ),
            ]
        ),
        "toc": ContainerNode(
            children=[
                TocNode(
                    children=[
                        TocEntryNode(
                            value=SentenceNode(children=[TextNode("Header 1")]),
                            anchor="Header 1-XXXXXX",
                            children=[
                                TocEntryNode(
                                    value=SentenceNode(
                                        children=[TextNode("Header 1.1")]
                                    ),
                                    anchor="Header 1.1-XXXXXX",
                                    children=[],
                                ),
                            ],
                        ),
                        TocEntryNode(
                            value=SentenceNode(children=[TextNode("Header 2")]),
                            anchor="Header 2-XXXXXX",
                            children=[],
                        ),
                    ]
                )
            ]
        ),
    }

    toc_node = parser.output["toc"].children[0]

    assert toc_node.parent == parser.output["toc"]


def test_toc_inside_block():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: f"{text}-XXXXXX"
    )

    source = """
    ----
    = Header 1
    == Header 1.1
    = Header 2

    ::toc:
    ----
    """

    parser = runner(source, environment)

    assert parser.output == {
        "content": ContainerNode(
            children=[
                BlockNode(
                    children=[
                        HeaderNode(
                            value=SentenceNode(children=[TextNode("Header 1")]),
                            level="1",
                            anchor="Header 1-XXXXXX",
                        ),
                        HeaderNode(
                            value=SentenceNode(children=[TextNode("Header 1.1")]),
                            level="2",
                            anchor="Header 1.1-XXXXXX",
                        ),
                        HeaderNode(
                            value=SentenceNode(children=[TextNode("Header 2")]),
                            level="1",
                            anchor="Header 2-XXXXXX",
                        ),
                        TocNode(
                            children=[
                                TocEntryNode(
                                    value=SentenceNode(children=[TextNode("Header 1")]),
                                    anchor="Header 1-XXXXXX",
                                    children=[
                                        TocEntryNode(
                                            value=SentenceNode(
                                                children=[TextNode("Header 1.1")]
                                            ),
                                            anchor="Header 1.1-XXXXXX",
                                            children=[],
                                        ),
                                    ],
                                ),
                                TocEntryNode(
                                    value=SentenceNode(children=[TextNode("Header 2")]),
                                    anchor="Header 2-XXXXXX",
                                    children=[],
                                ),
                            ]
                        ),
                    ],
                    preprocessor="none",
                )
            ]
        ),
        "toc": ContainerNode(
            children=[
                TocNode(
                    children=[
                        TocEntryNode(
                            value=SentenceNode(children=[TextNode("Header 1")]),
                            anchor="Header 1-XXXXXX",
                            children=[
                                TocEntryNode(
                                    value=SentenceNode(
                                        children=[TextNode("Header 1.1")]
                                    ),
                                    anchor="Header 1.1-XXXXXX",
                                    children=[],
                                ),
                            ],
                        ),
                        TocEntryNode(
                            value=SentenceNode(children=[TextNode("Header 2")]),
                            anchor="Header 2-XXXXXX",
                            children=[],
                        ),
                    ]
                )
            ]
        ),
    }

    toc_node = parser.output["toc"].children[0]

    assert toc_node.parent == parser.output["toc"]



from mau.environment.environment import Environment
from mau.nodes.content import ContentImageNode, ContentNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.visitors.base_visitor import BaseVisitor


def test_content_node():
    visitor = BaseVisitor(Environment())

    node = ContentNode(
        content_type="sometype",
        uris=["/uri1", "/uri2"],
        title=SentenceNode(children=[TextNode("sometitle")]),
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "content",
            "content_type": "sometype",
            "uris": ["/uri1", "/uri2"],
            "title": {
                "data": {
                    "content": [
                        {
                            "data": {
                                "type": "text",
                                "value": "sometitle",
                                "args": [],
                                "kwargs": {},
                                "subtype": None,
                                "tags": [],
                            }
                        },
                    ],
                    "type": "sentence",
                    "args": [],
                    "kwargs": {},
                    "subtype": None,
                    "tags": [],
                },
            },
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
            "subtype": None,
        }
    }


def test_content_image_node():
    visitor = BaseVisitor(Environment())

    node = ContentImageNode(
        uri="someuri",
        alt_text="sometext",
        classes=["class1", "class2"],
        title=SentenceNode(children=[TextNode("sometitle")]),
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "content_image",
            "uri": "someuri",
            "title": {
                "data": {
                    "content": [
                        {
                            "data": {
                                "type": "text",
                                "value": "sometitle",
                                "args": [],
                                "kwargs": {},
                                "subtype": None,
                                "tags": [],
                            }
                        },
                    ],
                    "type": "sentence",
                    "args": [],
                    "kwargs": {},
                    "subtype": None,
                    "tags": [],
                },
            },
            "alt_text": "sometext",
            "classes": ["class1", "class2"],
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
            "subtype": None,
        }
    }



from mau.lexers.main_lexer import MainLexer
from mau.nodes.content import ContentImageNode, ContentNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_include_content():
    source = """
    ["text", #tag1, key1=value1, key2=value2]
    << ctype1:/path/to/it,/another/path
    """

    assert runner(source).nodes == [
        ContentNode(
            uris=["/path/to/it", "/another/path"],
            content_type="ctype1",
            args=["text"],
            tags=["tag1"],
            kwargs={"key1": "value1", "key2": "value2"},
        )
    ]


def test_include_content_with_subtype():
    source = """
    [*subtype1]
    << ctype1:/path/to/it
    """

    assert runner(source).nodes == [
        ContentNode(
            uris=["/path/to/it"],
            content_type="ctype1",
            subtype="subtype1",
            args=[],
            tags=[],
            kwargs={},
        )
    ]


def test_include_image_with_only_path():
    source = """
    << image:/path/to/it.jpg
    """

    assert runner(source).nodes == [ContentImageNode("/path/to/it.jpg")]


def test_include_image_with_http():
    source = """
    << image:https:///some.domain/path/to/it.jpg
    """

    assert runner(source).nodes == [
        ContentImageNode("https:///some.domain/path/to/it.jpg")
    ]


def test_include_image_with_arguments():
    source = """
    ["alt text", #tag1, key1=value1, key2=value2, classes="class1,class2"]
    << image:/path/to/it.jpg,
    """

    assert runner(source).nodes == [
        ContentImageNode(
            "/path/to/it.jpg",
            args=[],
            tags=["tag1"],
            kwargs={"key1": "value1", "key2": "value2"},
            alt_text="alt text",
            classes=["class1", "class2"],
        )
    ]


def test_include_image_with_title():
    source = """
    . A nice caption
    << image:/path/to/it.jpg
    """

    assert runner(source).nodes == [
        ContentImageNode(
            "/path/to/it.jpg",
            title=SentenceNode(
                children=[
                    TextNode("A nice caption"),
                ]
            ),
        )
    ]


def test_include_image_with_subtype():
    source = """
    [*subtype1]
    << image:/path/to/it.jpg
    """

    assert runner(source).nodes == [
        ContentImageNode(
            "/path/to/it.jpg",
            subtype="subtype1",
        ),
    ]



from mau.environment.environment import Environment
from mau.nodes.header import HeaderNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.visitors.base_visitor import BaseVisitor


def test_header_node():
    visitor = BaseVisitor(Environment())

    node = HeaderNode(
        value=SentenceNode(children=[TextNode("Just some text")]),
        level="3",
        anchor="someanchor",
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "header",
            "value": {
                "data": {
                    "type": "sentence",
                    "content": [
                        {
                            "data": {
                                "type": "text",
                                "value": "Just some text",
                                "subtype": None,
                                "args": [],
                                "kwargs": {},
                                "tags": [],
                            },
                        }
                    ],
                    "subtype": None,
                    "args": [],
                    "kwargs": {},
                    "tags": [],
                }
            },
            "level": 3,
            "anchor": "someanchor",
            "subtype": None,
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
        }
    }



class AttributesManager:
    def __init__(self, parser):
        self.args = []
        self.kwargs = {}
        self.tags = []
        self.subtype = None

        self.parser = parser

    @property
    def attributes(self):
        return (
            self.args,
            self.kwargs,
            self.tags,
            self.subtype,
        )

    def push(
        self,
        args=None,
        kwargs=None,
        tags=None,
        subtype=None,
    ):
        self.args = args or self.args
        self.kwargs = kwargs or self.kwargs
        self.tags = tags or self.tags
        self.subtype = subtype or self.subtype

    def pop(self):
        args, kwargs, tags, subtype = self.attributes

        self.args = []
        self.kwargs = {}
        self.tags = []
        self.subtype = None

        return (args, kwargs, tags, subtype)



from mau.environment.environment import Environment
from mau.nodes.footnotes import FootnoteNode, FootnotesNode
from mau.nodes.inline import TextNode
from mau.visitors.jinja_visitor import JinjaVisitor


def test_page_footnotes_node():
    templates = {
        "text.j2": "{{ value }}",
        "footnotes.j2": "{{ entries }}",
        "footnotes_entry.j2": (
            "{{ content }}:{{ number }}:" "{{ content_anchor }}:{{ reference_anchor }}"
        ),
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = FootnotesNode(
        children=[
            FootnoteNode(
                children=[TextNode("Footnote 1")],
                number="1",
                reference_anchor="anchor-1",
                content_anchor="anchor-1-def",
            ).to_entry(),
            FootnoteNode(
                children=[TextNode("Footnote 2")],
                number="2",
                reference_anchor="anchor-2",
                content_anchor="anchor-2-def",
            ).to_entry(),
        ],
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert (
        result == "Footnote 1:1:anchor-1-def:anchor-1Footnote 2:2:anchor-2-def:anchor-2"
    )



from abc import ABC
from collections.abc import MutableMapping


class EnvironmentABC(ABC):
    pass


def flatten_nested_dict(nested, parent_key=None, separator="."):
    flat = {}

    for key, value in nested.items():
        key = f"{parent_key}{separator}{key}" if parent_key else key

        if isinstance(value, MutableMapping) and len(value) != 0:
            flat.update(flatten_nested_dict(value, key, separator=separator))
        else:
            flat[key] = value

    return flat


def nest_flattened_dict(flat, separator="."):
    def _split(key, value, separator, output):
        key, *rest = key.split(separator, 1)

        if rest:
            _split(
                rest[0],
                value,
                separator,
                output.setdefault(key, {}),
            )
        else:
            output[key] = value

    result = {}

    for key, value in flat.items():
        _split(key, value, separator, result)

    return result


class Environment(EnvironmentABC):
    def __init__(self, other=None):
        self._variables = {}

        if other is not None:
            self.update(other)

    def clone(self):
        return self.__class__(self._variables)

    def setvar(self, key, value):
        self.update({key: value})
        # self._variables[key] = value

    def getvar(self, key, default=None):
        try:
            return self._variables[key]
        except KeyError:
            prefix = f"{key}."

            keys = [k for k in self._variables if k.startswith(prefix)]

            if len(keys) != 0:
                return Environment(
                    {
                        k.removeprefix(prefix): v
                        for k, v in self._variables.items()
                        if k.startswith(prefix)
                    }
                )

            return default

    def getvar_nodefault(self, key):
        return self._variables[key]

    def update(self, other, namespace=None):
        if isinstance(other, EnvironmentABC):
            if not namespace:
                self._variables.update(other._variables)
                return

            self._variables.update(flatten_nested_dict({namespace: other._variables}))
        else:
            if not namespace:
                self._variables.update(flatten_nested_dict(other))
                return

            self._variables.update(flatten_nested_dict({namespace: other}))

    def asdict(self):
        return nest_flattened_dict(self._variables)

    def asflatdict(self):
        return self._variables

    def __repr__(self):  # pragma: no cover
        return f"{self.asdict()}"

    def __eq__(self, other):
        return self._variables == other._variables



from mau.environment.environment import Environment
from mau.nodes.footnotes import FootnoteNode
from mau.nodes.header import HeaderNode
from mau.nodes.inline import RawNode, SentenceNode, StyleNode, TextNode, VerbatimNode
from mau.nodes.macros import (
    MacroClassNode,
    MacroHeaderNode,
    MacroImageNode,
    MacroLinkNode,
    MacroNode,
)
from mau.nodes.nodes import Node
from mau.visitors.base_visitor import BaseVisitor


def test_no_node():
    visitor = BaseVisitor(Environment())

    result = visitor.visit(None)

    assert result == {}


def test_unknown_node():
    visitor = BaseVisitor(Environment())

    node = Node("Just some text.")
    node.node_type = "unknown"

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "unknown",
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_raw_node():
    visitor = BaseVisitor(Environment())

    node = RawNode("Just some text.")

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "raw",
            "value": "Just some text.",
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_text_node():
    visitor = BaseVisitor(Environment())

    node = TextNode("Just some text.")

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "text",
            "value": "Just some text.",
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_verbatim_node():
    visitor = BaseVisitor(Environment())

    node = VerbatimNode("Just some text.")

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "verbatim",
            "value": "Just some text.",
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_sentence_node():
    visitor = BaseVisitor(Environment())

    node = SentenceNode(
        children=[
            TextNode("Just some text. "),
            TextNode("More text."),
        ]
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "sentence",
            "content": [
                {
                    "data": {
                        "type": "text",
                        "value": "Just some text. ",
                        "subtype": None,
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                    }
                },
                {
                    "data": {
                        "type": "text",
                        "value": "More text.",
                        "subtype": None,
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                    }
                },
            ],
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_style_node_star():
    visitor = BaseVisitor(Environment())

    node = StyleNode("star", children=[TextNode("Just some text.")])

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "style",
            "content": [
                {
                    "data": {
                        "type": "text",
                        "value": "Just some text.",
                        "subtype": None,
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                    }
                },
            ],
            "value": "star",
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_style_node_underscore():
    visitor = BaseVisitor(Environment())

    node = StyleNode("underscore", children=[TextNode("Just some text.")])

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "style",
            "content": [
                {
                    "data": {
                        "type": "text",
                        "value": "Just some text.",
                        "subtype": None,
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                    }
                },
            ],
            "value": "underscore",
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_style_node_tilde():
    visitor = BaseVisitor(Environment())

    node = StyleNode("tilde", children=[TextNode("Just some text.")])

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "style",
            "content": [
                {
                    "data": {
                        "type": "text",
                        "value": "Just some text.",
                        "subtype": None,
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                    }
                },
            ],
            "value": "tilde",
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_style_node_caret():
    visitor = BaseVisitor(Environment())

    node = StyleNode("caret", children=[TextNode("Just some text.")])

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "style",
            "content": [
                {
                    "data": {
                        "type": "text",
                        "value": "Just some text.",
                        "subtype": None,
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                    }
                },
            ],
            "value": "caret",
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_macro_node():
    visitor = BaseVisitor(Environment())

    node = MacroNode(name="somename", args=["arg1", "arg2"], kwargs={"key1": "value1"})

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "macro",
            "name": "somename",
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": [],
            "subtype": None,
        }
    }


def test_footnote_node():
    visitor = BaseVisitor(Environment())

    node = FootnoteNode(
        children=[TextNode("Just some text.")],
        number="5",
        reference_anchor="someanchor",
        content_anchor="someanchor-def",
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "footnote",
            "number": "5",
            "reference_anchor": "someanchor",
            "content_anchor": "someanchor-def",
            "content": [
                {
                    "data": {
                        "type": "text",
                        "value": "Just some text.",
                        "subtype": None,
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                    }
                },
            ],
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_class_node():
    visitor = BaseVisitor(Environment())

    node = MacroClassNode(["class1", "class2"], children=[TextNode("Just some text.")])

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "macro.class",
            "content": [
                {
                    "data": {
                        "type": "text",
                        "value": "Just some text.",
                        "subtype": None,
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                    }
                }
            ],
            "classes": ["class1", "class2"],
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_link_node():
    visitor = BaseVisitor(Environment())

    node = MacroLinkNode(target="sometarget", children=[TextNode("sometext")])

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "macro.link",
            "content": [
                {
                    "data": {
                        "type": "text",
                        "value": "sometext",
                        "subtype": None,
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                    }
                }
            ],
            "target": "sometarget",
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_header_node():
    visitor = BaseVisitor(Environment())

    header_node = HeaderNode(
        value=SentenceNode(children=[TextNode("Header")]),
        level="2",
        anchor="XXXXXX",
        kwargs={"id": "someid"},
    )

    node = MacroHeaderNode(
        header_id="someid", header=header_node, children=[TextNode("sometext")]
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "macro.header",
            "content": [
                {
                    "data": {
                        "type": "text",
                        "value": "sometext",
                        "subtype": None,
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                    }
                }
            ],
            "header": {
                "value": {
                    "data": {
                        "type": "sentence",
                        "subtype": None,
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                        "content": [
                            {
                                "data": {
                                    "type": "text",
                                    "value": "Header",
                                    "subtype": None,
                                    "args": [],
                                    "kwargs": {},
                                    "tags": [],
                                }
                            }
                        ],
                    }
                },
                "level": "2",
                "anchor": "XXXXXX",
            },
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_header_node_with_no_text():
    visitor = BaseVisitor(Environment())

    header_node = HeaderNode(
        value=SentenceNode(children=[TextNode("Header")]),
        level="2",
        anchor="XXXXXX",
        kwargs={"id": "someid"},
    )

    node = MacroHeaderNode(header_id="someid", header=header_node, children=[])

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "macro.header",
            "content": {
                "data": {
                    "type": "sentence",
                    "subtype": None,
                    "args": [],
                    "kwargs": {},
                    "tags": [],
                    "content": [
                        {
                            "data": {
                                "type": "text",
                                "value": "Header",
                                "subtype": None,
                                "args": [],
                                "kwargs": {},
                                "tags": [],
                            }
                        }
                    ],
                }
            },
            "header": {
                "value": {
                    "data": {
                        "type": "sentence",
                        "subtype": None,
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                        "content": [
                            {
                                "data": {
                                    "type": "text",
                                    "value": "Header",
                                    "subtype": None,
                                    "args": [],
                                    "kwargs": {},
                                    "tags": [],
                                }
                            }
                        ],
                    }
                },
                "level": "2",
                "anchor": "XXXXXX",
            },
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }


def test_image_node():
    visitor = BaseVisitor(Environment())

    node = MacroImageNode(uri="someuri", alt_text="sometext", width="100", height="400")

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "macro.image",
            "uri": "someuri",
            "alt_text": "sometext",
            "width": "100",
            "height": "400",
            "subtype": None,
            "args": [],
            "kwargs": {},
            "tags": [],
        }
    }



from mau.lexers.main_lexer import MainLexer
from mau.nodes.block import BlockNode
from mau.nodes.inline import StyleNode, TextNode
from mau.nodes.paragraph import ParagraphNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_admonition():
    source = """
    [*admonition,someclass,someicon]
    ----
    Content
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="admonition",
            children=[
                ParagraphNode(
                    children=[
                        TextNode("Content"),
                    ]
                ),
            ],
            secondary_children=[],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={"class": "someclass", "icon": "someicon"},
        )
    ]


def test_parse_block_quote_attribution_in_secondary_content():
    source = """
    [*quote]
    ----
    Learn about the Force, Luke.
    ----
    _Star Wars_, 1977
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="quote",
            children=[
                ParagraphNode(
                    children=[
                        TextNode("Learn about the Force, Luke."),
                    ]
                ),
            ],
            secondary_children=[
                ParagraphNode(
                    children=[
                        StyleNode(
                            value="underscore",
                            children=[
                                TextNode("Star Wars"),
                            ],
                        ),
                        TextNode(", 1977"),
                    ]
                ),
            ],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]



import pytest
from mau.environment.environment import (
    Environment,
    flatten_nested_dict,
    nest_flattened_dict,
)


def test_flatten():
    nested = {
        "var1": {
            "var2": "value2",
            "var3": {
                "var4": "value4",
                "var5": "value5",
            },
        },
    }

    assert flatten_nested_dict(nested) == {
        "var1.var2": "value2",
        "var1.var3.var4": "value4",
        "var1.var3.var5": "value5",
    }


def test_flatten_empty():
    nested = {
        "var1": {
            "var2": {},
        },
    }

    assert flatten_nested_dict(nested) == {
        "var1.var2": {},
    }


def test_nest():
    flat = {
        "var1.var2": "value2",
        "var1.var3.var4": "value4",
        "var1.var3.var5": "value5",
    }

    assert nest_flattened_dict(flat) == {
        "var1": {
            "var2": "value2",
            "var3": {
                "var4": "value4",
                "var5": "value5",
            },
        },
    }


def test_nest_empty():
    flat = {
        "var1.var2": {},
    }

    assert nest_flattened_dict(flat) == {
        "var1": {
            "var2": {},
        },
    }


def test_init():
    environment = Environment()

    assert environment.asdict() == {}


def test_init_with_flat_content():
    environment = Environment({"var1": "value1", "var2": "value2"})

    assert environment.asdict() == {"var1": "value1", "var2": "value2"}


def test_init_with_nested_content():
    environment = Environment({"var1": {"var2": "value2"}})

    assert environment.asdict() == {"var1": {"var2": "value2"}}


def test_equality():
    assert Environment(
        {
            "var1": "value1",
            "var2": "value2",
        }
    ) == Environment(
        {
            "var1": "value1",
            "var2": "value2",
        }
    )


def test_update():
    environment = Environment()

    environment.update({"var1": "value1", "var2": "value2"})

    assert environment.asdict() == {"var1": "value1", "var2": "value2"}


def test_update_does_not_replace():
    environment = Environment({"var1": "value1", "var2": "value2"})

    environment.update({"var3": "value3"})

    assert environment.asdict() == {
        "var1": "value1",
        "var2": "value2",
        "var3": "value3",
    }


def test_update_with_namespace():
    environment = Environment()

    environment.update({"var1": "value1", "var2": "value2"}, "test")

    assert environment.asdict() == {"test": {"var1": "value1", "var2": "value2"}}


def test_update_with_namespace_does_not_replace():
    environment = Environment({"test": {"var1": "value1", "var2": "value2"}})

    environment.update({"var3": "value3"}, "test")

    assert environment.asdict() == {
        "test": {"var1": "value1", "var2": "value2", "var3": "value3"}
    }


def test_update_from_other_environment():
    environment = Environment()

    environment.update(Environment({"var1": "value1", "var2": "value2"}))

    assert environment.asdict() == {"var1": "value1", "var2": "value2"}


def test_update_with_namespace_from_other_environment():
    environment = Environment()

    environment.update(Environment({"var1": "value1", "var2": "value2"}), "test")

    assert environment.asdict() == {"test": {"var1": "value1", "var2": "value2"}}


def test_update_deep():
    environment = Environment({"mau": {"visitor": {"class": "someclass"}}})

    environment.update(
        {"visitor": {"custom_templates": {"template1": "value1"}}}, "mau"
    )

    assert environment.asdict() == {
        "mau": {
            "visitor": {
                "class": "someclass",
                "custom_templates": {"template1": "value1"},
            },
        },
    }


def test_clone():
    environment = Environment({"var1": "value1", "var2": "value2"})
    clone = environment.clone

    assert not clone is environment
    assert environment.clone().asdict() == {"var1": "value1", "var2": "value2"}


def test_get_variable_flat():
    environment = Environment({"var1": "value1"})

    assert environment.getvar("var1") == "value1"


def test_get_variable_nested():
    environment = Environment({"var1": {"var2": "value2"}})

    assert environment.getvar("var1.var2") == "value2"


def test_get_variable_nested_empty():
    environment = Environment({"var1": {"var2": {}}})

    assert environment.getvar("var1.var2") == {}


def test_get_variable_flat_default():
    environment = Environment({"var1": "value1"})

    assert environment.getvar("var2", "def") == "def"


def test_get_variable_nested_default():
    environment = Environment({"var1": {"var2": "value2"}})

    assert environment.getvar("var1.var3", "def") == "def"


def test_set_variable_flat():
    environment = Environment()

    environment.setvar("var1", "value1")

    assert environment._variables == {"var1": "value1"}


def test_set_variable_nested():
    environment = Environment()

    environment.setvar("var1.var2", "value2")

    assert environment._variables == {"var1.var2": "value2"}


def test_set_variable_dict():
    environment = Environment()

    environment.setvar("var1", {"var2": "value2"})

    assert environment._variables == {"var1.var2": "value2"}


def test_set_variable_empty_dict():
    environment = Environment()

    environment.setvar("var1", {})
    environment.setvar("var2", {"var3": {}})

    assert environment._variables == {
        "var1": {},
        "var2.var3": {},
    }
    assert environment.getvar("var1") == {}
    assert environment.getvar("var2.var3") == {}


def test_asdict():
    environment = Environment()

    environment.setvar("var1.var2", "value2")
    environment.setvar("var3.var4.var5", "value5")

    assert environment.asdict() == {
        "var1": {
            "var2": "value2",
        },
        "var3": {
            "var4": {
                "var5": "value5",
            },
        },
    }


def test_get_variable_flat_no_default():
    environment = Environment()

    with pytest.raises(KeyError):
        environment.getvar_nodefault("var1")


def test_get_variable_nested_no_default():
    environment = Environment()

    with pytest.raises(KeyError):
        environment.getvar_nodefault("var1.var2")


def test_get_variable_is_a_namespace():
    environment = Environment(
        {
            "top": {
                "middle": {
                    "var1": "value1",
                    "var2": "value2",
                    "var3": {},
                }
            },
        }
    )

    assert environment.getvar("top")._variables == {
        "middle.var1": "value1",
        "middle.var2": "value2",
        "middle.var3": {},
    }

    assert environment.getvar("top.middle")._variables == {
        "var1": "value1",
        "var2": "value2",
        "var3": {},
    }

    assert environment.getvar("notthere") is None

    assert environment.getvar("top.mid") is None



import textwrap

import pytest
from mau.environment.environment import Environment
from mau.errors import MauErrorException
from mau.lexers.main_lexer import MainLexer
from mau.nodes.header import HeaderNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.macros import MacroHeaderNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_internal_link():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: "XXXXXX"
    )

    source = """
    This is a paragraph with an internal link [header](someid, text).

    [id=someid]
    == Header
    """

    parser = runner(textwrap.dedent(source), environment)
    parser.finalise()

    header_node = HeaderNode(
        value=SentenceNode(children=[TextNode("Header")]),
        level="2",
        anchor="XXXXXX",
        kwargs={"id": "someid"},
    )

    assert parser.internal_links_manager.links == [
        MacroHeaderNode(
            header_id="someid", header=header_node, children=[TextNode("text")]
        )
    ]
    assert parser.internal_links_manager.headers == {"someid": header_node}


def test_internal_link_no_header():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: "XXXXXX"
    )

    source = """
    This is a paragraph with an internal link [header](someotherid, text).

    [id=someid]
    == Header
    """

    with pytest.raises(MauErrorException):
        runner(textwrap.dedent(source))


def test_internal_link_header_in_block():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: "XXXXXX"
    )

    source = """
    This is a paragraph with an internal link [header](someid, text).

    ====
    [id=someid]
    == Header
    ====
    """

    parser = runner(textwrap.dedent(source), environment)
    parser.finalise()

    header_node = HeaderNode(
        value=SentenceNode(children=[TextNode("Header")]),
        level="2",
        anchor="XXXXXX",
        kwargs={"id": "someid"},
    )

    assert parser.internal_links_manager.links == [
        MacroHeaderNode(
            header_id="someid", header=header_node, children=[TextNode("text")]
        )
    ]
    assert parser.internal_links_manager.headers == {"someid": header_node}


def test_internal_link_link_in_block():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: "XXXXXX"
    )

    source = """
    ====
    This is a paragraph with an internal link [header](someid, text).
    ====

    [id=someid]
    == Header
    """

    parser = runner(textwrap.dedent(source), environment)
    parser.finalise()

    header_node = HeaderNode(
        value=SentenceNode(children=[TextNode("Header")]),
        level="2",
        anchor="XXXXXX",
        kwargs={"id": "someid"},
    )

    assert parser.internal_links_manager.links == [
        MacroHeaderNode(
            header_id="someid", header=header_node, children=[TextNode("text")]
        )
    ]
    assert parser.internal_links_manager.headers == {"someid": header_node}


def test_internal_link_link_and_header_in_block():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: "XXXXXX"
    )

    source = """
    ====
    This is a paragraph with an internal link [header](someid, text).

    [id=someid]
    == Header
    ====
    """

    parser = runner(textwrap.dedent(source), environment)
    parser.finalise()

    header_node = HeaderNode(
        value=SentenceNode(children=[TextNode("Header")]),
        level="2",
        anchor="XXXXXX",
        kwargs={"id": "someid"},
    )

    assert parser.internal_links_manager.links == [
        MacroHeaderNode(
            header_id="someid", header=header_node, children=[TextNode("text")]
        )
    ]
    assert parser.internal_links_manager.headers == {"someid": header_node}



from mau.environment.environment import Environment
from mau.nodes.inline import TextNode
from mau.nodes.source import CalloutNode, CalloutsEntryNode, SourceNode
from mau.visitors.base_visitor import BaseVisitor


def test_source_node():
    visitor = BaseVisitor(Environment())

    node = SourceNode(
        language="somelang",
        code=[
            TextNode("import sys"),
            TextNode("import: os"),
            TextNode(""),
            TextNode('print(os.environ["HOME"])'),
        ],
        markers=[CalloutNode(1, "imp"), CalloutNode(3, "env")],
        callouts=[
            CalloutsEntryNode("imp", "This is an import"),
            CalloutsEntryNode("env", "Environment variables are paramount"),
        ],
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "source",
            "subtype": None,
            "callouts": [
                {
                    "data": {
                        "marker": "imp",
                        "type": "callouts_entry",
                        "value": "This is an import",
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                },
                {
                    "data": {
                        "marker": "env",
                        "type": "callouts_entry",
                        "value": "Environment variables are paramount",
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                },
            ],
            "classes": [],
            "code": [
                {
                    "data": {
                        "type": "text",
                        "value": "import sys",
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                },
                {
                    "data": {
                        "type": "text",
                        "value": "import: os",
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                },
                {
                    "data": {
                        "type": "text",
                        "value": "",
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                },
                {
                    "data": {
                        "type": "text",
                        "value": 'print(os.environ["HOME"])',
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                },
            ],
            "highlights": [],
            "language": "somelang",
            "lines": 4,
            "markers": [
                None,
                {
                    "data": {
                        "type": "callout",
                        "line": 1,
                        "marker": "imp",
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                },
                None,
                {
                    "data": {
                        "type": "callout",
                        "line": 3,
                        "marker": "env",
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                },
            ],
            "preprocessor": None,
            "title": {},
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
            "subtype": None,
        }
    }



from mau.nodes.inline import RawNode, StyleNode, TextNode, VerbatimNode, WordNode


def test_word_node():
    node = WordNode("somevalue")

    assert node.value == "somevalue"
    assert node.node_type == "word"
    assert node == WordNode("somevalue")


def test_text_node():
    node = TextNode("somevalue")

    assert node.value == "somevalue"
    assert node.node_type == "text"
    assert node == TextNode("somevalue")


def test_raw_node():
    node = RawNode("somevalue")

    assert node.value == "somevalue"
    assert node.node_type == "raw"
    assert node == RawNode("somevalue")


def test_verbatim_node():
    node = VerbatimNode("somevalue")

    assert node.value == "somevalue"
    assert node.node_type == "verbatim"
    assert node == VerbatimNode("somevalue")


def test_style_node():
    node = StyleNode("mystyle", children=[TextNode("othervalue")])

    assert node.value == "mystyle"
    assert node.children == [TextNode("othervalue")]
    assert node.node_type == "style"
    assert node == StyleNode("mystyle", children=[TextNode("othervalue")])



from mau.environment.environment import Environment
from mau.nodes.toc import TocEntryNode, TocNode
from mau.visitors.jinja_visitor import JinjaVisitor


def test_page_toc_node():
    templates = {
        "text.j2": "{{ value }}",
        "toc.j2": (
            "{{ entries }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
        "toc_entry.j2": (
            "{{ anchor }}:{{ value }}{% if children %} - " "{{ children }}{% endif %}"
        ),
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = TocNode(
        entries=[
            TocEntryNode(
                "Header 1",
                "header-1",
                children=[
                    TocEntryNode("Header 1.1", "header-1-1"),
                ],
            ),
            TocEntryNode("Header 2", "header-2"),
        ],
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == (
        "header-1:Header 1 - header-1-1:Header 1.1header-2:Header 2 - "
        "arg1,arg2 - key1:value1 - tag1,tag2"
    )



from unittest.mock import patch

import pytest
from mau.environment.environment import Environment
from mau.errors import MauErrorException
from mau.lexers.main_lexer import MainLexer
from mau.nodes.block import BlockNode
from mau.nodes.header import HeaderNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_block_attributes_args_and_kwargs():
    source = """
    [*subtype,arg1,key1=value1]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="subtype",
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=["arg1"],
            kwargs={"key1": "value1"},
        )
    ]


def test_block_attributes_can_contain_variables():
    source = """
    :value1:42

    [*subtype,key1={value1}]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="subtype",
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={"key1": "42"},
        )
    ]


def test_parse_block_title_and_attributes():
    source = """
    .Just a title
    [*subtype, name1=value1, name2=value2]
    ----
    ----
    """

    parser = runner(source)

    assert parser.nodes == [
        BlockNode(
            subtype="subtype",
            classes=[],
            title=SentenceNode(
                children=[
                    TextNode("Just a title"),
                ]
            ),
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={"name1": "value1", "name2": "value2"},
        )
    ]

    block_node = parser.nodes[0]
    text_node = block_node.title.children[0]

    assert text_node.parent == block_node
    assert text_node.parent_position == "title"


def test_parse_block_title_and_attributes_are_reset():
    source = """
    .Just a title
    [*subtype1, name1=value1, name2=value2]
    ----
    ----

    [*subtype2]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="subtype1",
            classes=[],
            title=SentenceNode(
                children=[
                    TextNode("Just a title"),
                ]
            ),
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={"name1": "value1", "name2": "value2"},
        ),
        BlockNode(
            subtype="subtype2",
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={},
        ),
    ]


def test_block_classes_single_class():
    source = """
    [*subtype,classes=cls1]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="subtype",
            classes=["cls1"],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={},
        ),
    ]


def test_block_classes_multiple_classes():
    source = """
    [*subtype,classes="cls1,cls2"]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="subtype",
            classes=["cls1", "cls2"],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={},
        ),
    ]


def test_block_engine():
    source = """
    [*subtype,engine=someengine]
    ----
    ----
    """

    with pytest.raises(MauErrorException):
        runner(source)


@patch("mau.parsers.main_parser.header_anchor")
def test_block_default_engine_adds_headers_to_global_toc(mock_header_anchor):
    mock_header_anchor.return_value = "XXYY"

    source = """
    = Global header

    [*someblock]
    ----
    = Block header
    ----
    """

    par = runner(source)

    assert par.nodes == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Global header")]),
            level="1",
            anchor="XXYY",
        ),
        BlockNode(
            subtype="someblock",
            children=[
                HeaderNode(
                    value=SentenceNode(children=[TextNode("Block header")]),
                    level="1",
                    anchor="XXYY",
                ),
            ],
            secondary_children=[],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={},
        ),
    ]

    assert par.toc_manager.headers == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Global header")]),
            level="1",
            anchor="XXYY",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Block header")]),
            level="1",
            anchor="XXYY",
        ),
    ]


def test_command_defblock():
    source = """
    ::defblock:alias, *myblock, language=python, engine=source
    """

    par = runner(source)

    assert par.block_aliases["alias"] == {
        "subtype": "myblock",
        "mandatory_args": [],
        "defaults": {"language": "python", "engine": "source"},
    }


def test_define_block_in_environment():
    source = """
    """

    environment = Environment()
    environment.setvar(
        "mau.parser.block_definitions",
        {
            "alias": {
                "subtype": "type",
                "mandatory_args": ["arg1", "arg2"],
                "defaults": {"key1": "value1"},
            },
        },
    )

    par = runner(source, environment=environment)

    assert par.block_aliases["alias"] == {
        "subtype": "type",
        "mandatory_args": ["arg1", "arg2"],
        "defaults": {"key1": "value1"},
    }


def test_command_defblock_no_subtype():
    source = """
    ::defblock:alias, language=python, engine=source
    """

    par = runner(source)

    assert par.block_aliases["alias"] == {
        "subtype": None,
        "mandatory_args": [],
        "defaults": {"language": "python", "engine": "source"},
    }


def test_command_defblock_source():
    source = """
    """

    par = runner(source)

    assert par.block_aliases["source"] == {
        "subtype": None,
        "mandatory_args": ["language"],
        "defaults": {"language": "text", "engine": "source"},
    }


def test_command_defblock_source_can_be_overridden():
    source = """
    ::defblock:source, *source, language=python, engine=source
    """

    par = runner(source)

    assert par.block_aliases["source"] == {
        "subtype": "source",
        "mandatory_args": [],
        "defaults": {"language": "python", "engine": "source"},
    }


def test_command_defblock_no_args():
    source = """
    ::defblock:
    """

    with pytest.raises(MauErrorException):
        runner(source)


def test_block_definitions_are_used():
    source = """
    ::defblock:alias, *subtype, name1=value1, name2=value2

    [*alias]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="subtype",
            children=[],
            secondary_children=[],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={"name1": "value1", "name2": "value2"},
        ),
    ]


def test_block_definitions_local_kwargs_overwrite_defined_ones():
    source = """
    ::defblock:alias, *subtype, name1=value1, name2=value2

    [*alias, name1=value99]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="subtype",
            children=[],
            secondary_children=[],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={"name1": "value99", "name2": "value2"},
        ),
    ]


def test_block_definitions_local_args_are_used():
    # This is testing that local args are used even when
    # an alias is used. However, the test should fail as
    # blocks cannot have unnamed arguments.

    source = """
    ::defblock:alias, *subtype, name1=value1, name2=value2

    [*alias, attr1, name1=value99]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="subtype",
            children=[],
            secondary_children=[],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=["attr1"],
            kwargs={"name1": "value99", "name2": "value2"},
        ),
    ]


def test_block_definitions_unnamed_args_are_used_as_names():
    # This is testing that the unnamed args specified
    # in the block definition are used as names for the
    # unnamed arguments passed to the block.

    source = """
    ::defblock:alias, *subtype, attr1, attr2

    [*alias, value1, value2]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="subtype",
            children=[],
            secondary_children=[],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={"attr1": "value1", "attr2": "value2"},
        ),
    ]


def test_block_definitions_unnamed_and_named_args():
    # This is testing that any named are specifid in
    # the block definition is included in the block
    # kwargs if not redefined.

    source = """
    ::defblock:alias, *subtype, attr1, attr2, attr3=value3

    [*alias, value1, value2]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="subtype",
            children=[],
            secondary_children=[],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={"attr1": "value1", "attr2": "value2", "attr3": "value3"},
        ),
    ]


def test_block_definitions_no_values_for_unnamed_args():
    # This is testing that is there are unnamed args in
    # the block definiton they should be given values
    # when the block is created.

    source = """
    ::defblock:alias, *subtype, attr1, attr2, attr3=value3

    [*alias]
    ----
    ----
    """

    with pytest.raises(ValueError):
        runner(source)


def test_block_definitions_values_without_unnamed_args():
    # This is testing that values passed to the
    # block at creation time should match the parameters
    # specified in the block definition.

    source = """
    ::defblock:alias, *subtype, attr3=value3

    [*alias, value1, value2]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="subtype",
            children=[],
            secondary_children=[],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=["value1", "value2"],
            kwargs={"attr3": "value3"},
        ),
    ]


def test_block_definitions_named_override_unnamed():
    # This is testing that values passed to the
    # block at creation time should match the parameters
    # specified in the block definition.

    source = """
    ::defblock:alias, *subtype, arg1, key1=value1

    [*alias, value2, arg1=value1]
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="subtype",
            children=[],
            secondary_children=[],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=["value2"],
            kwargs={"arg1": "value1", "key1": "value1"},
        ),
    ]



from mau.lexers.text_lexer import TextLexer
from mau.nodes.inline import TextNode
from mau.nodes.macros import MacroHeaderNode
from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(TextLexer, TextParser)

runner = parser_runner_factory(TextLexer, TextParser)


def test_macro_header():
    source = '[header](id, "link text")'

    node = MacroHeaderNode("id", children=[TextNode("link text")])

    parser = runner(source)
    assert parser.nodes == [node]
    assert parser.links == [node]


def test_macro_header_without_text():
    source = "[header](id)"

    node = MacroHeaderNode("id", children=[])

    parser = runner(source)
    assert parser.nodes == [node]
    assert parser.links == [node]



from unittest.mock import patch

import pytest
from mau.errors import MauErrorException
from mau.lexers.main_lexer import MainLexer
from mau.nodes.block import BlockNode
from mau.nodes.header import HeaderNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.toc import TocEntryNode, TocNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


@patch("mau.parsers.main_parser.header_anchor")
def test_engine_mau(mock_header_anchor):
    mock_header_anchor.return_value = "XXYY"

    source = """
    = Header 1

    = Header 2

    [*sometype, engine=mau]
    ----
    = Header 3

    = Header 4
    ----
    """

    par = runner(source)

    assert par.nodes == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 1")]),
            level="1",
            anchor="XXYY",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 2")]),
            level="1",
            anchor="XXYY",
        ),
        BlockNode(
            subtype="sometype",
            children=[
                HeaderNode(
                    value=SentenceNode(children=[TextNode("Header 3")]),
                    level="1",
                    anchor="XXYY",
                ),
                HeaderNode(
                    value=SentenceNode(children=[TextNode("Header 4")]),
                    level="1",
                    anchor="XXYY",
                ),
            ],
            secondary_children=[],
            classes=[],
            title=None,
            engine="mau",
            preprocessor="none",
            args=[],
            kwargs={},
        ),
    ]

    assert par.toc_manager.headers == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 1")]),
            level="1",
            anchor="XXYY",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 2")]),
            level="1",
            anchor="XXYY",
        ),
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_engine_mau_multiple_blocks_are_independent(mock_header_anchor):
    mock_header_anchor.return_value = "XXYY"

    source = """
    = Header 1

    = Header 2

    [*sometype, engine=mau]
    ----
    = Header 3

    = Header 4
    ----

    [*sometype, engine=mau]
    ----
    = Header 5

    = Header 6
    ----
    """

    par = runner(source)

    assert par.nodes == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 1")]),
            level="1",
            anchor="XXYY",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 2")]),
            level="1",
            anchor="XXYY",
        ),
        BlockNode(
            subtype="sometype",
            children=[
                HeaderNode(
                    value=SentenceNode(children=[TextNode("Header 3")]),
                    level="1",
                    anchor="XXYY",
                ),
                HeaderNode(
                    value=SentenceNode(children=[TextNode("Header 4")]),
                    level="1",
                    anchor="XXYY",
                ),
            ],
            secondary_children=[],
            classes=[],
            title=None,
            engine="mau",
            preprocessor="none",
            args=[],
            kwargs={},
        ),
        BlockNode(
            subtype="sometype",
            children=[
                HeaderNode(
                    value=SentenceNode(children=[TextNode("Header 5")]),
                    level="1",
                    anchor="XXYY",
                ),
                HeaderNode(
                    value=SentenceNode(children=[TextNode("Header 6")]),
                    level="1",
                    anchor="XXYY",
                ),
            ],
            secondary_children=[],
            classes=[],
            title=None,
            engine="mau",
            preprocessor="none",
            args=[],
            kwargs={},
        ),
    ]

    assert par.toc_manager.headers == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 1")]),
            level="1",
            anchor="XXYY",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 2")]),
            level="1",
            anchor="XXYY",
        ),
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_engine_mau_toc(mock_header_anchor):
    mock_header_anchor.return_value = "XXYY"

    source = """
    = Header 1

    = Header 2

    [*sometype, engine=mau]
    ----
    = Header 3

    = Header 4

    ::toc:
    ----

    ::toc:
    """

    par = runner(source)

    assert par.nodes == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 1")]),
            level="1",
            anchor="XXYY",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 2")]),
            level="1",
            anchor="XXYY",
        ),
        BlockNode(
            subtype="sometype",
            children=[
                HeaderNode(
                    value=SentenceNode(children=[TextNode("Header 3")]),
                    level="1",
                    anchor="XXYY",
                ),
                HeaderNode(
                    value=SentenceNode(children=[TextNode("Header 4")]),
                    level="1",
                    anchor="XXYY",
                ),
                TocNode(
                    children=[
                        TocEntryNode(
                            value=SentenceNode(children=[TextNode("Header 3")]),
                            anchor="XXYY",
                        ),
                        TocEntryNode(
                            value=SentenceNode(children=[TextNode("Header 4")]),
                            anchor="XXYY",
                        ),
                    ]
                ),
            ],
            secondary_children=[],
            classes=[],
            title=None,
            engine="mau",
            preprocessor="none",
            args=[],
            kwargs={},
        ),
        TocNode(
            children=[
                TocEntryNode(
                    value=SentenceNode(children=[TextNode("Header 1")]), anchor="XXYY"
                ),
                TocEntryNode(
                    value=SentenceNode(children=[TextNode("Header 2")]), anchor="XXYY"
                ),
            ]
        ),
    ]

    assert par.toc_manager.headers == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 1")]),
            level="1",
            anchor="XXYY",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 2")]),
            level="1",
            anchor="XXYY",
        ),
    ]


def test_block_mau_has_no_external_variables():
    source = """
    :answer:42
    [*block, engine=mau]
    ----
    The answer is {answer}.
    ----
    """

    with pytest.raises(MauErrorException):
        assert runner(source)



from mau.nodes.header import HeaderNode
from mau.nodes.inline import TextNode
from mau.nodes.macros import (
    MacroClassNode,
    MacroHeaderNode,
    MacroImageNode,
    MacroLinkNode,
    MacroNode,
)


def test_macro_node():
    node = MacroNode("aname", args=["value"], kwargs={"akey": "avalue"})

    assert node.name == "aname"
    assert node.args == ["value"]
    assert node.kwargs == {"akey": "avalue"}
    assert node.node_type == "macro"
    assert node == MacroNode("aname", args=["value"], kwargs={"akey": "avalue"})


def test_class_node():
    node = MacroClassNode(["class1", "class2"], children=[TextNode("othervalue")])

    assert node.classes == ["class1", "class2"]
    assert node.children == [TextNode("othervalue")]
    assert node.node_type == "macro.class"
    assert node == MacroClassNode(
        ["class1", "class2"], children=[TextNode("othervalue")]
    )


def test_link_node():
    node = MacroLinkNode("atarget", children=[TextNode("sometext")])

    assert node.target == "atarget"
    assert node.children == [TextNode("sometext")]
    assert node.node_type == "macro.link"
    assert node == MacroLinkNode("atarget", children=[TextNode("sometext")])


def test_link_node_no_text():
    node = MacroLinkNode("atarget")

    assert node.target == "atarget"
    assert node.children == []
    assert node.node_type == "macro.link"
    assert node == MacroLinkNode("atarget")


def test_image_node():
    node = MacroImageNode("someuri", "somealttext", "width", "height")

    assert node.uri == "someuri"
    assert node.alt_text == "somealttext"
    assert node.width == "width"
    assert node.height == "height"
    assert node.node_type == "macro.image"
    assert node == MacroImageNode("someuri", "somealttext", "width", "height")


def test_header_node():
    header_node = HeaderNode("someheader", "1", "someanchor")
    node = MacroHeaderNode("someid", header_node, children=[TextNode("sometext")])

    assert node.header_id == "someid"
    assert node.header == header_node
    assert node.children == [TextNode("sometext")]
    assert node.node_type == "macro.header"
    assert node == MacroHeaderNode(
        "someid", header_node, children=[TextNode("sometext")]
    )



from mau.nodes.content import ContentImageNode, ContentNode


def test_content_node():
    node = ContentNode(
        content_type="somecontent", uris=["/uri1", "/uri2"], title="sometitle"
    )

    assert node.content_type == "somecontent"
    assert node.uris == ["/uri1", "/uri2"]
    assert node.title == "sometitle"
    assert node.args == []
    assert node.kwargs == {}
    assert node.node_type == "content"
    assert node == ContentNode(
        content_type="somecontent", uris=["/uri1", "/uri2"], title="sometitle"
    )


def test_content_image_node():
    node = ContentImageNode(
        uri="someuri", alt_text="somealttext", classes=[], title="sometitle"
    )

    assert node.uri == "someuri"
    assert node.alt_text == "somealttext"
    assert node.classes == []
    assert node.title == "sometitle"
    assert node.args == []
    assert node.kwargs == {}
    assert node.node_type == "content_image"
    assert node == ContentImageNode(
        uri="someuri", alt_text="somealttext", classes=[], title="sometitle"
    )



from mau.environment.environment import Environment
from mau.nodes.inline import TextNode
from mau.nodes.nodes import Node
from mau.nodes.paragraph import ParagraphNode
from mau.visitors.jinja_visitor import JinjaVisitor, create_templates


def test_create_templates_prefixes():
    child_node = Node()
    child_node.node_type = "node_type"

    parent_node = Node()
    parent_node.node_type = "parent_type"

    child_node.parent = parent_node
    parent_node.children.append(child_node)

    templates = create_templates(
        ["prefix1", "prefix2"], ["node_template1"], child_node, "ext"
    )

    assert templates == [
        "prefix1.parent_type.node_template1.ext",
        "prefix1.parent_type.node_type.ext",
        "prefix1.node_template1.ext",
        "prefix1.node_type.ext",
        "prefix2.parent_type.node_template1.ext",
        "prefix2.parent_type.node_type.ext",
        "prefix2.node_template1.ext",
        "prefix2.node_type.ext",
        "parent_type.node_template1.ext",
        "parent_type.node_type.ext",
        "node_template1.ext",
        "node_type.ext",
    ]


def test_create_templates_node_subtype():
    child_node = Node()
    child_node.node_type = "node_type"
    child_node.subtype = "node_subtype"

    parent_node = Node()
    parent_node.node_type = "parent_type"

    child_node.parent = parent_node
    parent_node.children.append(child_node)

    templates = create_templates([], ["node_template1"], child_node, "ext")

    assert templates == [
        "parent_type.node_template1.node_subtype.ext",
        "parent_type.node_template1.ext",
        "parent_type.node_type.node_subtype.ext",
        "parent_type.node_type.ext",
        "node_template1.node_subtype.ext",
        "node_template1.ext",
        "node_type.node_subtype.ext",
        "node_type.ext",
    ]


def test_create_templates_node_templates():
    child_node = Node()
    child_node.node_type = "node_type"

    parent_node = Node()
    parent_node.node_type = "parent_type"

    child_node.parent = parent_node
    parent_node.children.append(child_node)

    templates = create_templates(
        [], ["node_template1", "node_template2"], child_node, "ext"
    )

    assert templates == [
        "parent_type.node_template1.ext",
        "parent_type.node_template2.ext",
        "parent_type.node_type.ext",
        "node_template1.ext",
        "node_template2.ext",
        "node_type.ext",
    ]


def test_create_templates_parent_subtype():
    child_node = Node()
    child_node.node_type = "node_type"

    parent_node = Node()
    parent_node.node_type = "parent_type"
    parent_node.subtype = "parent_subtype"

    child_node.parent = parent_node
    parent_node.children.append(child_node)

    templates = create_templates([], ["node_template1"], child_node, "ext")

    assert templates == [
        "parent_type.parent_subtype.node_template1.ext",
        "parent_type.parent_subtype.node_type.ext",
        "parent_type.node_template1.ext",
        "parent_type.node_type.ext",
        "node_template1.ext",
        "node_type.ext",
    ]


def test_create_templates_parent_position():
    child_node = Node()
    child_node.node_type = "node_type"

    parent_node = Node()
    parent_node.node_type = "parent_type"
    parent_node.subtype = "parent_subtype"

    child_node.parent = parent_node
    child_node.parent_position = "primary"
    parent_node.children.append(child_node)

    templates = create_templates([], ["node_template1"], child_node, "ext")

    assert templates == [
        "parent_type.parent_subtype.primary.node_template1.ext",
        "parent_type.parent_subtype.primary.node_type.ext",
        "parent_type.parent_subtype.node_template1.ext",
        "parent_type.parent_subtype.node_type.ext",
        "parent_type.primary.node_template1.ext",
        "parent_type.primary.node_type.ext",
        "parent_type.node_template1.ext",
        "parent_type.node_type.ext",
        "node_template1.ext",
        "node_type.ext",
    ]


def test_create_templates():
    child_node = Node()
    child_node.node_type = "node_type"
    child_node.subtype = "node_subtype"
    child_node.tags = ["tag1", "tag2"]

    templates = create_templates([], ["node_template1"], child_node, "ext")

    assert templates == [
        "node_template1.node_subtype.tag1.ext",
        "node_template1.node_subtype.tag2.ext",
        "node_template1.node_subtype.ext",
        "node_template1.tag1.ext",
        "node_template1.tag2.ext",
        "node_template1.ext",
        "node_type.node_subtype.tag1.ext",
        "node_type.node_subtype.tag2.ext",
        "node_type.node_subtype.ext",
        "node_type.tag1.ext",
        "node_type.tag2.ext",
        "node_type.ext",
    ]


def test_inline_text_node_without_prefix():
    templates = {
        "text.j2": "{{ value }}",
        "prefix.text.j2": "PREFIXED {{ value }}",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    environment.setvar("mau.visitor.prefixes", [])
    visitor = JinjaVisitor(environment)

    node = TextNode("Just some text.")

    result = visitor.visit(node)

    assert result == "Just some text."


def test_inline_text_node_with_prefix():
    templates = {
        "text.j2": "{{ value }}",
        "prefix.text.j2": "PREFIXED {{ value }}",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    environment.setvar("mau.visitor.prefixes", ["prefix"])
    visitor = JinjaVisitor(environment)

    node = TextNode("Just some text.")

    result = visitor.visit(node)

    assert result == "PREFIXED Just some text."


def test_inline_text_node_with_multiple_prefixes():
    templates = {
        "text.j2": "{{ value }}",
        "otherprefix.text.j2": "PREFIXED OTHER {{ value }}",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    environment.setvar("mau.visitor.prefixes", ["prefix", "otherprefix"])
    visitor = JinjaVisitor(environment)

    node = TextNode("Just some text.")

    result = visitor.visit(node)

    assert result == "PREFIXED OTHER Just some text."


def test_inline_text_node_with_parent():
    templates = {
        "text.j2": "{{ value }}",
        "paragraph.text.j2": "PARAGRAPH {{ value }}",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    text_node = TextNode("Just some text.")
    paragraph_node = ParagraphNode()
    text_node.parent = paragraph_node
    paragraph_node.children = [text_node]

    result = visitor.visit(text_node)

    assert result == "PARAGRAPH Just some text."


def test_inline_text_node_with_parent_subtype():
    templates = {
        "text.j2": "{{ value }}",
        "paragraph.text.j2": "PARAGRAPH {{ value }}",
        "paragraph.fancy.text.j2": "FANCY PARAGRAPH {{ value }}",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    text_node = TextNode("Just some text.")
    paragraph_node = ParagraphNode(subtype="fancy")
    text_node.parent = paragraph_node
    paragraph_node.children = [text_node]

    result = visitor.visit(text_node)

    assert result == "FANCY PARAGRAPH Just some text."
